import glob
import os
import pathlib
import re
import runpy
import shutil
import sys
import tempfile
import urllib.request
import zipfile

import pyxel


def cli():
    num_args = len(sys.argv)
    command = sys.argv[1] if num_args > 1 else ""
    if command == "run" and num_args == 3:
        run_python_script(sys.argv[2])
    elif command == "play" and num_args == 3:
        play_pyxel_app(sys.argv[2])
    elif command == "edit" and (num_args == 2 or num_args == 3):
        edit_pyxel_resource(sys.argv[2] if num_args == 3 else None)
    elif command == "package" and num_args == 4:
        package_pyxel_app(sys.argv[2], sys.argv[3])
    elif command == "copy_examples" and num_args == 2:
        copy_pyxel_examples()
    else:
        _print_usage()


def _print_usage():
    print(f"Pyxel {pyxel.PYXEL_VERSION}, a retro game engine for Python")
    print("usage:")
    print("    pyxel run PYTHON_SCRIPT_FILE(.py)")
    print("    pyxel play PYXEL_APP_FILE(.pyxapp)")
    print("    pyxel edit [PYXEL_RESOURCE_FILE(.pyxres)]")
    print("    pyxel package APP_ROOT_DIR STARTUP_SCRIPT_FILE(.py)")
    print("    pyxel copy_examples")
    _check_newer_version()


def _check_newer_version():
    url = "https://www.github.com/kitao/pyxel"
    req = urllib.request.Request(url)
    latest_version = None
    try:
        with urllib.request.urlopen(req, timeout=3) as res:
            pattern = r"/kitao/pyxel/releases/tag/v(\d+\.\d+\.\d+)"
            text = res.read().decode("utf-8")
            result = re.search(pattern, text)
            if result:
                latest_version = result.group(1)
    except urllib.error.URLError:
        return
    if not latest_version:
        return

    def parse_version(version):
        return list(map(int, version.split(".")))

    if parse_version(latest_version) > parse_version(pyxel.PYXEL_VERSION):
        print(f"A new version, Pyxel {latest_version}, is available.")


def _complete_extension(filename, ext_with_dot):
    file_ext = os.path.splitext(filename)[1]
    if file_ext == ext_with_dot:
        return filename
    else:
        return filename + ext_with_dot


def _files_in_dir(dirname):
    paths = glob.glob(os.path.join(dirname, "**/*"), recursive=True)
    return sorted(list(filter(os.path.isfile, paths)))


def _check_file_exists(filename):
    if not os.path.isfile(filename):
        print(f"no such file: '{filename}'")
        exit(1)


def _check_dir_exists(dirname):
    if not os.path.isdir(dirname):
        print(f"no such directory: '{dirname}'")
        exit(1)


def _make_app_dir():
    play_dir = os.path.join(tempfile.gettempdir(), pyxel.PYXEL_WORKING_DIR, "play")
    pathlib.Path(play_dir).mkdir(parents=True, exist_ok=True)
    for path in glob.glob(os.path.join(play_dir, "*")):
        pid = int(os.path.basename(path))
        if not pyxel.process_exists(pid):
            shutil.rmtree(path)
    app_dir = os.path.join(play_dir, str(os.getpid()))
    os.mkdir(app_dir)
    return app_dir


def run_python_script(python_script_file):
    python_script_file = _complete_extension(python_script_file, ".py")
    _check_file_exists(python_script_file)
    sys.path.append(os.path.dirname(python_script_file))
    runpy.run_path(python_script_file)


def play_pyxel_app(pyxel_app_file):
    pyxel_app_file = _complete_extension(pyxel_app_file, pyxel.APP_FILE_EXTENSION)
    _check_file_exists(pyxel_app_file)
    app_dir = _make_app_dir()
    zf = zipfile.ZipFile(pyxel_app_file)
    zf.extractall(app_dir)
    pattern = os.path.join(app_dir, "*", pyxel.APP_STARTUP_SCRIPT_FILE)
    for setting_file in glob.glob(pattern):
        with open(setting_file, "r") as f:
            startup_script_file = os.path.join(os.path.dirname(setting_file), f.read())
            sys.path.append(os.path.dirname(startup_script_file))
            runpy.run_path(startup_script_file)
            return
    print(f"file not found: '{pyxel.APP_STARTUP_SCRIPT_FILE}'")
    exit(1)


def edit_pyxel_resource(pyxel_resource_file):
    import pyxel.editor

    pyxel_resource_file = pyxel_resource_file or "my_resource"
    pyxel_resource_file = _complete_extension(
        pyxel_resource_file, pyxel.RESOURCE_FILE_EXTENSION
    )
    pyxel.editor.App(pyxel_resource_file)


def package_pyxel_app(app_root_dir, startup_script_name):
    _check_dir_exists(app_root_dir)
    startup_script_name = _complete_extension(startup_script_name, ".py")
    _check_file_exists(os.path.join(app_root_dir, startup_script_name))
    setting_file = os.path.join(app_root_dir, pyxel.APP_STARTUP_SCRIPT_FILE)
    with open(setting_file, "w") as f:
        f.write(startup_script_name)
    pyxel_app_file = os.path.basename(app_root_dir) + pyxel.APP_FILE_EXTENSION
    app_parent_dir = os.path.dirname(os.path.abspath(app_root_dir))
    with zipfile.ZipFile(
        pyxel_app_file,
        "w",
        compression=zipfile.ZIP_DEFLATED,
    ) as zf:
        files = [setting_file] + _files_in_dir(app_root_dir)
        for file in files:
            arcname = os.path.relpath(file, app_parent_dir)
            zf.write(file, arcname)
            print(f"added '{arcname}'")
    os.remove(setting_file)


def copy_pyxel_examples():
    src_dir = os.path.join(os.path.dirname(__file__), "examples")
    dst_dir = "pyxel_examples"
    shutil.rmtree(dst_dir, ignore_errors=True)
    for src_file in _files_in_dir(src_dir):
        dst_file = os.path.join(dst_dir, os.path.relpath(src_file, src_dir))
        os.makedirs(os.path.dirname(dst_file), exist_ok=True)
        shutil.copyfile(src_file, dst_file)
        print(f"copied '{dst_file}'")



import pyxel

from .settings import (
    EDITOR_IMAGE,
    MAX_SOUND_LENGTH,
    SOUND_FIELD_CURSOR_EDIT_COLOR,
    SOUND_FIELD_CURSOR_SELECT_COLOR,
    SOUND_FIELD_DATA_NORMAL_COLOR,
    SOUND_FIELD_DATA_SELECT_COLOR,
    TEXT_LABEL_COLOR,
)
from .widgets import Widget
from .widgets.settings import WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME

TONE_KEY_TABLE = [pyxel.KEY_T, pyxel.KEY_S, pyxel.KEY_P, pyxel.KEY_N]
EFFECT_KEY_TABLE = [pyxel.KEY_N, pyxel.KEY_S, pyxel.KEY_V, pyxel.KEY_F]


class SoundField(Widget):
    """
    Variables:
        is_playing_var
        help_message_var
    """

    def __init__(self, parent):
        super().__init__(parent, 30, 149, 193, 23)
        self.field_cursor = parent.field_cursor
        self.get_field = parent.get_field
        self.get_field_help_message = parent.get_field_help_message
        self.copy_var("is_playing_var", parent)
        self.copy_var("help_message_var", parent)

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_hover", self.__on_mouse_hover)
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

    def _screen_to_view(self, x, y):
        x = min(max((x - self.x - 1) // 4, 0), MAX_SOUND_LENGTH - 1)
        y = min(max((y - self.y) // 8, 0), 2)
        return x, y

    def __on_mouse_down(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT or self.is_playing_var:
            return
        x, y = self._screen_to_view(x, y)
        self.field_cursor.move_to(x, y + 1, pyxel.btn(pyxel.KEY_SHIFT))

    def __on_mouse_hover(self, x, y):
        self.help_message_var = self.get_field_help_message()

    def __on_update(self):
        cursor_y = self.field_cursor.y
        if (
            cursor_y < 1
            or self.is_playing_var
            or pyxel.btn(pyxel.KEY_SHIFT)
            or pyxel.btn(pyxel.KEY_CTRL)
            or pyxel.btn(pyxel.KEY_ALT)
            or pyxel.btn(pyxel.KEY_GUI)
        ):
            return
        value = None
        if cursor_y == 1:
            for i in range(4):
                if pyxel.btnp(TONE_KEY_TABLE[i], WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
                    value = i
                    break
        elif cursor_y == 2:
            for i in range(8):
                key = pyxel.KEY_0 if i == 0 else pyxel.KEY_1 + i - 1
                if pyxel.btnp(key, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME) or pyxel.btnp(
                    key, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME
                ):
                    value = i
                    break
        elif cursor_y == 3:
            for i in range(4):
                if pyxel.btnp(
                    EFFECT_KEY_TABLE[i], WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME
                ):
                    value = i
                    break
        if value is None:
            return
        self.field_cursor.insert(value)

    def __on_draw(self):
        # Draw field frame
        pyxel.text(self.x - 13, self.y + 1, "TON", TEXT_LABEL_COLOR)
        pyxel.text(self.x - 13, self.y + 9, "VOL", TEXT_LABEL_COLOR)
        pyxel.text(self.x - 13, self.y + 17, "EFX", TEXT_LABEL_COLOR)
        pyxel.blt(
            self.x,
            self.y,
            EDITOR_IMAGE,
            0,
            79,
            193,
            23,
        )

        # Draw field data
        data_str = []
        data_str.append("".join(["TSPN"[v] for v in self.get_field(1)]))
        data_str.append("".join([str(v) for v in self.get_field(2)]))
        data_str.append("".join(["NSVF"[v] for v in self.get_field(3)]))
        for i in range(3):
            pyxel.text(31, 150 + i * 8, data_str[i], SOUND_FIELD_DATA_NORMAL_COLOR)

        # Draw cursor
        cursor_y = self.field_cursor.y
        cursor_x = self.field_cursor.x
        if self.is_playing_var or cursor_y == 0:
            return
        x = cursor_x * 4 + 31
        y = cursor_y * 8 + 142
        w = self.field_cursor.width * 4
        col = (
            SOUND_FIELD_CURSOR_SELECT_COLOR
            if self.field_cursor.is_selecting
            else SOUND_FIELD_CURSOR_EDIT_COLOR
        )
        pyxel.rect(x, y - 1, w, 7, col)
        if cursor_x < len(data_str[cursor_y - 1]):
            pyxel.text(
                x,
                y,
                data_str[cursor_y - 1][cursor_x : cursor_x + self.field_cursor.width],
                SOUND_FIELD_DATA_SELECT_COLOR,
            )



import pyxel

from .settings import EDITOR_IMAGE
from .widgets import Widget
from .widgets.settings import BUTTON_ENABLED_COLOR, BUTTON_PRESSED_COLOR


class SoundSelector(Widget):
    """
    Variables:
        is_playing_var
        help_message_var
    """

    def __init__(self, parent):
        super().__init__(parent, 11, 129, 218, 44)
        self._pressed_sound = None
        self._preview_sound = None
        self._last_preview_sound = None
        self.field_cursor = parent.field_cursor
        self.copy_var("is_playing_var", parent)
        self.copy_var("help_message_var", parent)

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_up", self.__on_mouse_up)
        self.add_event_listener("mouse_repeat", self.__on_mouse_down)
        self.add_event_listener("mouse_hover", self.__on_mouse_hover)
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

    def _hit_sound_button(self, x, y):
        x -= self.x + 6
        y -= self.y + 5
        if x < 0 or y < 0 or x > 205 or y > 33 or x % 13 > 10 or y % 9 > 6:
            return None
        return (y // 9) * 16 + x // 13

    def __on_mouse_down(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT or self.is_playing_var:
            return
        self._pressed_sound = self._hit_sound_button(x, y)
        if self._pressed_sound is not None:
            self.field_cursor.insert(self._pressed_sound)

    def __on_mouse_up(self, key, x, y):
        if key == pyxel.MOUSE_BUTTON_LEFT:
            self._pressed_sound = None

    def __on_mouse_hover(self, x, y):
        self.help_message_var = "PREVIEW:HOVER INSERT:CLICK"

    def __on_update(self):
        if self.is_playing_var:
            return
        mx = pyxel.mouse_x
        my = pyxel.mouse_y
        if self.is_hit(mx, my):
            self._preview_sound = self._hit_sound_button(mx, my)
            if (
                self._preview_sound is not None
                and self._preview_sound != self._last_preview_sound
            ):
                pyxel.play(0, self._preview_sound, loop=True)
        else:
            self._preview_sound = None
        if self._preview_sound is None and pyxel.play_pos(0) is not None:
            pyxel.stop(0)
        self._last_preview_sound = self._preview_sound

    def __on_draw(self):
        self.draw_panel(self.x, self.y, self.width, self.height)
        pyxel.blt(self.x + 6, self.y + 5, EDITOR_IMAGE, 0, 121, 206, 34)
        for i in range(pyxel.NUM_SOUNDS):
            if pyxel.sound(i).notes:
                self._draw_sound_button(i, BUTTON_ENABLED_COLOR)
        if self._pressed_sound is not None:
            self._draw_sound_button(self._pressed_sound, BUTTON_PRESSED_COLOR)

    def _draw_sound_button(self, snd, col):
        pyxel.pal(13, col)
        x = (snd % 16) * 13
        y = (snd // 16) * 9
        pyxel.blt(
            self.x + x + 6,
            self.y + y + 5,
            EDITOR_IMAGE,
            x,
            y + 121,
            11,
            7,
        )
        pyxel.pal()



import pyxel

from .settings import PANEL_FOCUS_BORDER_COLOR, PANEL_FOCUS_COLOR
from .widgets import ScrollBar, Widget


class ImageViewer(Widget):
    def __init__(self, parent):
        """
        Variables:
            image_no_var
            tilemap_no_var
            focus_x_var
            focus_y_var
            viewport_x_var
            viewport_y_var
            help_message_var
        """

        if hasattr(parent, "tilemap_no_var"):
            y = 80
            height = 66
            slider_amount = 8
            self._is_tilemap_mode = True
            self.copy_var("tilemap_no_var", parent)
        else:
            y = 16
            height = 130
            slider_amount = 16
            self._is_tilemap_mode = False
        super().__init__(parent, 157, y, 66, height)
        self._press_x = 0
        self._press_y = 0
        self._drag_offset_x = 0
        self._drag_offset_y = 0
        self.copy_var("image_no_var", parent)
        self.copy_var("help_message_var", parent)

        # Initialize focus_x_var
        self.new_var("focus_x_var", 0)
        self.add_var_event_listener("focus_x_var", "set", self.__on_focus_x_set)
        self.add_var_event_listener("focus_x_var", "change", self.__on_focus_x_change)

        # Initialize focus_y_var
        self.new_var("focus_y_var", 0)
        self.add_var_event_listener("focus_y_var", "set", self.__on_focus_y_set)
        self.add_var_event_listener("focus_y_var", "change", self.__on_focus_y_change)

        # Initialize focus_w_var
        self.new_var("focus_w_var", 1 if self._is_tilemap_mode else 2)

        # Initialize focus_h_var
        self.new_var("focus_h_var", 1 if self._is_tilemap_mode else 2)

        # Initialize horizontal scroll bar
        self._h_scroll_bar = ScrollBar(
            self,
            0,
            height - 1,
            width=66,
            scroll_amount=32,
            slider_amount=8,
            value=0,
        )
        self.copy_var("viewport_x_var", self._h_scroll_bar, "value_var")

        # Initialize vertical scroll bar
        self._v_scroll_bar = ScrollBar(
            self,
            65,
            0,
            height=height,
            scroll_amount=32,
            slider_amount=slider_amount,
            value=0,
        )
        self.copy_var("viewport_y_var", self._v_scroll_bar, "value_var")

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_drag", self.__on_mouse_drag)
        self.add_event_listener("mouse_hover", self.__on_mouse_hover)
        self.add_event_listener("draw", self.__on_draw)

    def _screen_to_focus(self, x, y):
        x = min(max(self.viewport_x_var + (x - self.x - 1) // 8, 0), 31)
        y = min(max(self.viewport_y_var + (y - self.y - 1) // 8, 0), 31)
        return x, y

    def __on_focus_x_set(self, value):
        return min(max(value, 0), 32 - self.focus_w_var)

    def __on_focus_x_change(self, value):
        fx = self.focus_x_var
        fw = self.focus_w_var
        vx = self.viewport_x_var
        vw = 8
        self.viewport_x_var += min(fx - vx, 0) + max(fx + fw - vx - vw, 0)

    def __on_focus_y_set(self, value):
        return min(max(value, 0), 32 - self.focus_h_var)

    def __on_focus_y_change(self, value):
        fy = self.focus_y_var
        fh = self.focus_h_var
        vy = self.viewport_y_var
        vh = 8 if self._is_tilemap_mode else 16
        self.viewport_y_var += min(fy - vy, 0) + max(fy + fh - vy - vh, 0)

    def __on_mouse_down(self, key, x, y):
        if key == pyxel.MOUSE_BUTTON_LEFT:
            self.focus_x_var, self.focus_y_var = self._screen_to_focus(x, y)
            self._press_x = self.focus_x_var
            self._press_y = self.focus_y_var
        elif key == pyxel.MOUSE_BUTTON_RIGHT:
            self._drag_offset_x = 0
            self._drag_offset_y = 0

    def __on_mouse_drag(self, key, x, y, dx, dy):
        if key == pyxel.MOUSE_BUTTON_LEFT:
            if self._is_tilemap_mode:
                last_focus_x = self.focus_x_var
                last_focus_y = self.focus_y_var
                self._focus_x_var, self._focus_y_var = self._screen_to_focus(x, y)
                self.focus_w_var = min(abs(self._focus_x_var - self._press_x) + 1, 8)
                self.focus_h_var = min(abs(self._focus_y_var - self._press_y) + 1, 8)
                self.focus_x_var = min(self._focus_x_var, last_focus_x)
                self.focus_y_var = min(self._focus_y_var, last_focus_y)
            else:
                self.__on_mouse_down(key, x, y)
        elif key == pyxel.MOUSE_BUTTON_RIGHT:
            self._drag_offset_x -= dx
            self._drag_offset_y -= dy
            if abs(self._drag_offset_x) >= 8:
                offset = self._drag_offset_x // 8
                self.viewport_x_var += offset
                self._drag_offset_x -= offset * 8
            if abs(self._drag_offset_y) >= 8:
                offset = self._drag_offset_y // 8
                self.viewport_y_var += offset
                self._drag_offset_y -= offset * 8

    def __on_mouse_hover(self, x, y):
        x, y = self._screen_to_focus(x, y)
        s = (
            "TILE:SHIFT+CURSOR"
            if self._is_tilemap_mode
            else "TARGET:CURSOR IMPORT:DROP"
        )
        self.help_message_var = s + f" ({x * 8},{y * 8})"

    def __on_draw(self):
        self.draw_panel(self.x, self.y, self.width, self.height)

        # Draw image
        pyxel.blt(
            self.x + 1,
            self.y + 1,
            self.image_no_var,
            self.viewport_x_var * 8,
            self.viewport_y_var * 8,
            self.width - 2,
            self.height - 2,
        )

        # Draw focus
        x = self.x + (self.focus_x_var - self.viewport_x_var) * 8 + 1
        y = self.y + (self.focus_y_var - self.viewport_y_var) * 8 + 1
        w = self.focus_w_var * 8
        h = self.focus_h_var * 8
        pyxel.clip(self.x + 1, self.y + 1, self.width - 2, self.height - 2)
        pyxel.rectb(x, y, w, h, PANEL_FOCUS_COLOR)
        pyxel.rectb(x + 1, y + 1, w - 2, h - 2, PANEL_FOCUS_BORDER_COLOR)
        pyxel.rectb(x - 1, y - 1, w + 2, h + 2, PANEL_FOCUS_BORDER_COLOR)
        pyxel.clip()



import pyxel

from .settings import (
    PANEL_SELECT_BORDER_COLOR,
    PANEL_SELECT_FRAME_COLOR,
    TOOL_BUCKET,
    TOOL_CIRC,
    TOOL_CIRCB,
    TOOL_PENCIL,
    TOOL_RECT,
    TOOL_RECTB,
    TOOL_SELECT,
)
from .widgets import ScrollBar, Widget
from .widgets.settings import WIDGET_HOLD_TIME, WIDGET_PANEL_COLOR, WIDGET_REPEAT_TIME


class CanvasPanel(Widget):
    """
    Variables:
        color_var
        tool_var
        image_no_var
        canvas_var
        focus_x_var
        focus_y_var
        help_message_var

        tilemap_no_var
        tile_x_var
        tile_y_var
        tile_w_var
        tile_h_var
    """

    def __init__(self, parent):
        super().__init__(parent, 11, 16, 130, 130)
        if hasattr(parent, "tilemap_no_var"):
            self._is_tilemap_mode = True
            self.copy_var("tilemap_no_var", parent)
            self.copy_var("tile_x_var", parent)
            self.copy_var("tile_y_var", parent)
            self.copy_var("tile_w_var", parent)
            self.copy_var("tile_h_var", parent)
        else:
            self._is_tilemap_mode = False
        self._history_data = None
        self._press_x = 0
        self._press_y = 0
        self._last_x = 0
        self._last_y = 0
        self._drag_offset_x = 0
        self._drag_offset_y = 0
        self._select_x1 = 0
        self._select_y1 = 0
        self._select_x2 = 0
        self._select_y2 = 0
        self._copy_buffer = None
        self._is_dragged = False
        self._is_assist_mode = False
        self._edit_canvas = (
            pyxel.Tilemap(16, 16, 0) if self._is_tilemap_mode else pyxel.Image(16, 16)
        )
        self.add_history = parent.add_history
        self.copy_var("color_var", parent)
        self.copy_var("tool_var", parent)
        self.copy_var("image_no_var", parent)
        self.copy_var("canvas_var", parent)
        self.copy_var("focus_x_var", parent)
        self.copy_var("focus_y_var", parent)
        self.copy_var("help_message_var", parent)

        # Initialize horizontal scroll bar
        self._h_scroll_bar = ScrollBar(
            self,
            0,
            129,
            width=130,
            scroll_amount=32,
            slider_amount=2,
            value=0,
        )
        self._h_scroll_bar.add_event_listener("change", self.__on_h_scroll_bar_change)
        self.add_var_event_listener("focus_x_var", "change", self.__on_focus_x_change)

        # Initialize vertical scroll bar
        self._v_scroll_bar = ScrollBar(
            self,
            129,
            0,
            height=130,
            scroll_amount=32,
            slider_amount=2,
            value=0,
        )
        self._v_scroll_bar.add_event_listener("change", self.__on_v_scroll_bar_change)
        self.add_var_event_listener("focus_y_var", "change", self.__on_focus_y_change)

        # Initialize event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_up", self.__on_mouse_up)
        self.add_event_listener("mouse_drag", self.__on_mouse_drag)
        self.add_event_listener("mouse_hover", self.__on_mouse_hover)
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

    def _screen_to_focus(self, x, y):
        x = min(max((x - self.x - 1) // 8, 0), 15)
        y = min(max((y - self.y - 1) // 8, 0), 15)
        return x, y

    def _add_pre_history(self):
        self._history_data = data = {}
        if self._is_tilemap_mode:
            data["tilemap_no"] = self.tilemap_no_var
        else:
            data["image_no"] = self.image_no_var
        data["focus_pos"] = (self.focus_x_var, self.focus_y_var)
        data["old_canvas"] = self.canvas_var.get_slice(
            self.focus_x_var * 8, self.focus_y_var * 8, 16, 16
        )

    def _add_post_history(self):
        data = self._history_data
        data["new_canvas"] = self.canvas_var.get_slice(
            self.focus_x_var * 8, self.focus_y_var * 8, 16, 16
        )
        if data["old_canvas"] != data["new_canvas"]:
            self.add_history(data)

    def _reset_edit_canvas(self):
        self._edit_canvas.blt(
            0,
            0,
            self.canvas_var,
            self.focus_x_var * 8,
            self.focus_y_var * 8,
            16,
            16,
        )
        if self._is_tilemap_mode:
            self._edit_canvas.refimg = self.canvas_var.refimg

    def _finish_edit_canvas(self):
        if not self._is_tilemap_mode:
            return
        for y in range(16):
            for x in range(16):
                if self._edit_canvas.pget(x, y) != (255, 255):
                    continue
                tile = (
                    self.tile_x_var + (x - self._press_x) % self.tile_w_var,
                    self.tile_y_var + (y - self._press_y) % self.tile_h_var,
                )
                self._edit_canvas.pset(x, y, tile)

    def __on_h_scroll_bar_change(self, value):
        self.focus_x_var = value

    def __on_v_scroll_bar_change(self, value):
        self.focus_y_var = value

    def __on_focus_x_change(self, value):
        self._h_scroll_bar.value_var = value

    def __on_focus_y_change(self, value):
        self._v_scroll_bar.value_var = value

    def __on_mouse_down(self, key, x, y):
        if key == pyxel.MOUSE_BUTTON_RIGHT:
            x = self.focus_x_var * 8 + (x - self.x) // 8
            y = self.focus_y_var * 8 + (y - self.y) // 8
            if self._is_tilemap_mode:
                (self.tile_x_var, self.tile_y_var) = self.canvas_var.pget(x, y)
            else:
                self.color_var = self.canvas_var.pget(x, y)
            return
        elif key != pyxel.MOUSE_BUTTON_LEFT:
            return
        x, y = self._screen_to_focus(x, y)
        self._press_x = self._last_x = x
        self._press_y = self._last_y = y
        self._is_dragged = True
        self._is_assist_mode = False
        if self.tool_var == TOOL_SELECT:
            self._reset_edit_canvas()
            self._select_x1 = self._select_x2 = x
            self._select_y1 = self._select_y2 = y
        elif self.tool_var >= TOOL_PENCIL and self.tool_var <= TOOL_CIRC:
            self._reset_edit_canvas()
            self._edit_canvas.pset(x, y, self.color_var)
            self._finish_edit_canvas()
        elif self.tool_var == TOOL_BUCKET:
            self._add_pre_history()
            self._reset_edit_canvas()
            self._edit_canvas.fill(x, y, self.color_var)
            self._finish_edit_canvas()
            self.canvas_var.blt(
                self.focus_x_var * 8,
                self.focus_y_var * 8,
                self._edit_canvas,
                0,
                0,
                16,
                16,
            )
            self._add_post_history()

    def __on_mouse_up(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT:
            return
        self._is_dragged = False
        if TOOL_PENCIL <= self.tool_var <= TOOL_CIRC:
            self._add_pre_history()
            self.canvas_var.blt(
                self.focus_x_var * 8,
                self.focus_y_var * 8,
                self._edit_canvas,
                0,
                0,
                16,
                16,
            )
            self._add_post_history()

    def __on_mouse_drag(self, key, x, y, dx, dy):
        if key == pyxel.MOUSE_BUTTON_LEFT:
            x1 = self._press_x
            y1 = self._press_y
            x2 = (x - self.x - 1) // 8
            y2 = (y - self.y - 1) // 8

            if TOOL_RECTB <= self.tool_var <= TOOL_CIRC and self._is_assist_mode:
                dx = x2 - x1
                dy = y2 - y1
                if abs(dx) > abs(dy):
                    y2 = y1 + abs(dx) * (1 if dy > 0 else -1)
                else:
                    x2 = x1 + abs(dy) * (1 if dx > 0 else -1)

            if self.tool_var == TOOL_SELECT:
                x2 = min(max(x2, 0), 15)
                y2 = min(max(y2, 0), 15)
                self._select_x1, self._select_x2 = (x1, x2) if x1 < x2 else (x2, x1)
                self._select_y1, self._select_y2 = (y1, y2) if y1 < y2 else (y2, y1)

            elif self.tool_var == TOOL_PENCIL:
                if self._is_assist_mode:
                    self._reset_edit_canvas()
                    self._edit_canvas.line(x1, y1, x2, y2, self.color_var)
                    self._finish_edit_canvas()
                else:
                    self._edit_canvas.line(
                        self._last_x, self._last_y, x2, y2, self.color_var
                    )
                    self._finish_edit_canvas()

            elif self.tool_var == TOOL_RECTB:
                self._reset_edit_canvas()
                self._edit_canvas.rectb2(
                    x1,
                    y1,
                    x2,
                    y2,
                    self.color_var,
                )
                self._finish_edit_canvas()

            elif self.tool_var == TOOL_RECT:
                self._reset_edit_canvas()
                self._edit_canvas.rect2(
                    x1,
                    y1,
                    x2,
                    y2,
                    self.color_var,
                )
                self._finish_edit_canvas()

            elif self.tool_var == TOOL_CIRCB:
                self._reset_edit_canvas()
                self._edit_canvas.ellib2(x1, y1, x2, y2, self.color_var)
                self._finish_edit_canvas()

            elif self.tool_var == TOOL_CIRC:
                self._reset_edit_canvas()
                self._edit_canvas.elli2(x1, y1, x2, y2, self.color_var)
                self._finish_edit_canvas()

            self._last_x = x2
            self._last_y = y2

        elif key == pyxel.MOUSE_BUTTON_RIGHT:
            self._drag_offset_x -= dx
            self._drag_offset_y -= dy
            if abs(self._drag_offset_x) >= 16:
                offset = self._drag_offset_x // 16
                self.focus_x_var += offset
                self._drag_offset_x -= offset * 16
            if abs(self._drag_offset_y) >= 16:
                offset = self._drag_offset_y // 16
                self.focus_y_var += offset
                self._drag_offset_y -= offset * 16

    def __on_mouse_hover(self, x, y):
        if self.tool_var == TOOL_SELECT:
            s = "COPY:CTRL+C/X/V FLIP:H/V"
        elif self._is_dragged:
            s = "ASSIST:SHIFT"
        else:
            s = "PICK:R-CLICK VIEW:R-DRAG"
        x, y = self._screen_to_focus(x, y)
        x += self.focus_x_var * 8
        y += self.focus_y_var * 8
        self.help_message_var = s + f" ({x},{y})"

    def __on_update(self):
        if self._is_dragged and not self._is_assist_mode and pyxel.btn(pyxel.KEY_SHIFT):
            self._is_assist_mode = True
            self.__on_mouse_drag(
                pyxel.MOUSE_BUTTON_LEFT, pyxel.mouse_x, pyxel.mouse_y, 0, 0
            )

        if self.tool_var == TOOL_SELECT and (
            pyxel.btn(pyxel.KEY_CTRL) or pyxel.btn(pyxel.KEY_GUI)
        ):
            # Ctrl+A: Select all
            if pyxel.btnp(pyxel.KEY_A):
                self._select_x1 = self._select_y1 = 0
                self._select_x2 = self._select_y2 = 15

            # Ctrl+C: Copy
            if pyxel.btnp(pyxel.KEY_C):
                self._copy_buffer = self.canvas_var.get_slice(
                    self.focus_x_var * 8 + self._select_x1,
                    self.focus_y_var * 8 + self._select_y1,
                    self._select_x2 - self._select_x1 + 1,
                    self._select_y2 - self._select_y1 + 1,
                )

            # Ctrl+X: Cut
            if pyxel.btnp(pyxel.KEY_X):
                x = self.focus_x_var * 8 + self._select_x1
                y = self.focus_y_var * 8 + self._select_y1
                w = self._select_x2 - self._select_x1 + 1
                h = self._select_y2 - self._select_y1 + 1
                self._copy_buffer = self.canvas_var.get_slice(x, y, w, h)
                self._add_pre_history()
                self.canvas_var.rect(x, y, w, h, (0, 0) if self._is_tilemap_mode else 0)
                self._add_post_history()

            # Ctrl+V: Paste
            if self._copy_buffer is not None and pyxel.btnp(pyxel.KEY_V):
                self._add_pre_history()
                width = len(self._copy_buffer[0])
                height = len(self._copy_buffer)
                width -= max(self._select_x1 + width - 16, 0)
                height -= max(self._select_y1 + height - 16, 0)
                self.canvas_var.set_slice(
                    self.focus_x_var * 8 + self._select_x1,
                    self.focus_y_var * 8 + self._select_y1,
                    self._copy_buffer,
                )
                self._add_post_history()

        if self.tool_var == TOOL_SELECT:
            # H: Flip horizontal
            if pyxel.btnp(pyxel.KEY_H):
                x = self.focus_x_var * 8 + self._select_x1
                y = self.focus_y_var * 8 + self._select_y1
                w = self._select_x2 - self._select_x1 + 1
                h = self._select_y2 - self._select_y1 + 1
                self._add_pre_history()
                self.canvas_var.blt(x, y, self.canvas_var, x, y, -w, h)
                self._add_post_history()

            # V: Flip vertical
            if pyxel.btnp(pyxel.KEY_V):
                x = self.focus_x_var * 8 + self._select_x1
                y = self.focus_y_var * 8 + self._select_y1
                w = self._select_x2 - self._select_x1 + 1
                h = self._select_y2 - self._select_y1 + 1
                self._add_pre_history()
                self.canvas_var.blt(x, y, self.canvas_var, x, y, w, -h)
                self._add_post_history()

        # Move tile focus
        if self._is_tilemap_mode and pyxel.btn(pyxel.KEY_SHIFT):
            if pyxel.btnp(pyxel.KEY_LEFT, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
                self.tile_x_var -= 1
            if pyxel.btnp(pyxel.KEY_RIGHT, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
                self.tile_x_var += 1
            if pyxel.btnp(pyxel.KEY_UP, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
                self.tile_y_var -= 1
            if pyxel.btnp(pyxel.KEY_DOWN, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
                self.tile_y_var += 1

        # Move target focus
        if not (
            pyxel.btn(pyxel.KEY_SHIFT)
            or pyxel.btn(pyxel.KEY_CTRL)
            or pyxel.btn(pyxel.KEY_ALT)
            or pyxel.btn(pyxel.KEY_GUI)
        ):
            if pyxel.btnp(pyxel.KEY_LEFT, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
                self.focus_x_var -= 1
            if pyxel.btnp(pyxel.KEY_RIGHT, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
                self.focus_x_var += 1
            if pyxel.btnp(pyxel.KEY_UP, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
                self.focus_y_var -= 1
            if pyxel.btnp(pyxel.KEY_DOWN, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
                self.focus_y_var += 1

    def __on_draw(self):
        self.draw_panel(self.x, self.y, self.width, self.height)

        # Draw edit panel
        canvas, offset_x, offset_y = (
            (self._edit_canvas, 0, 0)
            if self._is_dragged
            else (self.canvas_var, self.focus_x_var * 8, self.focus_y_var * 8)
        )
        if self._is_tilemap_mode:
            pyxel.bltm(
                self.x + 1,
                self.y + 1,
                canvas,
                offset_x * 8,
                offset_y * 8,
                128,
                128,
            )
        else:
            for yi in range(16):
                for xi in range(16):
                    pyxel.rect(
                        self.x + xi * 8 + 1,
                        self.y + yi * 8 + 1,
                        8,
                        8,
                        canvas.pget(offset_x + xi, offset_y + yi),
                    )
        pyxel.line(
            self.x + 1, self.y + 64, self.x + 128, self.y + 64, WIDGET_PANEL_COLOR
        )
        pyxel.line(
            self.x + 64, self.y + 1, self.x + 64, self.y + 128, WIDGET_PANEL_COLOR
        )

        # Draw selection area
        if self.tool_var == TOOL_SELECT and self._select_x1 >= 0:
            x = self._select_x1 * 8 + 12
            y = self._select_y1 * 8 + 17
            w = self._select_x2 * 8 - x + 20
            h = self._select_y2 * 8 - y + 25
            pyxel.clip(self.x + 1, self.y + 1, 128, 128)
            pyxel.rectb(x, y, w, h, PANEL_SELECT_FRAME_COLOR)
            pyxel.rectb(x + 1, y + 1, w - 2, h - 2, PANEL_SELECT_BORDER_COLOR)
            pyxel.rectb(x - 1, y - 1, w + 2, h + 2, PANEL_SELECT_BORDER_COLOR)
            pyxel.clip()



import pyxel


class App:
    def __init__(self):
        pyxel.init(200, 150, title="Pyxel Draw API")
        pyxel.mouse(True)
        pyxel.image(0).load(0, 0, "assets/cat_16x16.png")
        pyxel.image(1).load(0, 0, "assets/tileset_24x32.png")
        pyxel.tilemap(0).set(
            0,
            0,
            [
                "0201 0000 0200 0400 0100 0000 0003 0103 0203 0000 0002",
                "0202 0300 0001 0101 0201 0300 0000 0100 0200 0300 0003",
            ],
        )
        pyxel.tilemap(0).image = pyxel.image(1)
        self.pal_test_is_enabled = False
        self.clip_test_is_enabled = False
        pyxel.run(self.update, self.draw)

    def update(self):
        if pyxel.btnp(pyxel.KEY_Q):
            pyxel.quit()

        self.pal_test_is_enabled = (pyxel.frame_count // 30) % 10 >= 5
        self.clip_test_is_enabled = pyxel.btn(pyxel.KEY_SPACE)

    def draw(self):
        self.test_pal1()
        self.test_cls(6, 6)
        self.test_clip()
        self.test_pset(6, 20)
        self.test_line(106, 6)
        self.test_rect(6, 38)
        self.test_rectb(106, 38)
        self.test_circ(6, 61)
        self.test_circb(106, 61)
        self.test_blt(6, 88)
        self.test_bltm(106, 88)
        self.test_text(6, 124)
        self.test_pal2(106, 124)

    def test_pal1(self):
        if self.pal_test_is_enabled:
            pyxel.pal(5, 2)
            pyxel.pal(12, 7)
            pyxel.pal(7, 10)

    def test_pal2(self, x, y):
        pyxel.text(x, y, "pal(col1,col2)", 12)
        pyxel.pal()

    def test_cls(self, x, y):
        pyxel.cls(5)
        pyxel.text(x, y, "cls(col)", 7)

    def test_clip(self):
        pyxel.clip()
        if not self.clip_test_is_enabled:
            return
        x = pyxel.sin(pyxel.frame_count * 1.14) * 39 + 40
        y = pyxel.sin(pyxel.frame_count * 1.71) * 29 + 30
        w = 120
        h = 90
        pyxel.text(x, y - 8, "clip(x,y,w,h)", 14)
        pyxel.rectb(x - 1, y - 1, w + 2, h + 2, 14)
        pyxel.clip(x, y, w, h)

    def test_pset(self, x, y):
        pyxel.text(x, y, "pset(x,y,col)", 7)
        x += 4
        y += 10
        for i in range(16):
            pyxel.pset(x + i * 2, y, i)

    def test_line(self, x, y):
        pyxel.text(x, y, "line(x1,y1,x2,y2,col)", 7)
        x += 4
        y += 9
        col = 5
        for i in range(3):
            pyxel.line(x, y + i * 8, x + 48, y + i * 8, col)
            col += 1
        for i in range(4):
            pyxel.line(x + i * 16, y, x + i * 16, y + 16, col)
            col += 1
        for i in range(4):
            pyxel.line(x + i * 16, y, x + (3 - i) * 16, y + 16, col)
            col += 1

    def test_rect(self, x, y):
        pyxel.text(x, y, "rect(x,y,w,h,col)", 7)
        x += 4
        y += 16
        for i in range(8):
            pyxel.rect(x + i * 8, y - i, i + 1, i + 1, i + 8)

    def test_rectb(self, x, y):
        pyxel.text(x, y, "rectb(x,y,w,h,col)", 7)
        x += 4
        y += 16
        for i in range(8):
            pyxel.rectb(x + i * 8, y - i, i + 1, i + 1, i + 8)

    def test_circ(self, x, y):
        pyxel.text(x, y, "circ(x,y,r,col)", 7)
        x += 4
        y += 15
        for i in range(8):
            pyxel.circ(x + i * 8, y, i, i + 8)

    def test_circb(self, x, y):
        pyxel.text(x, y, "circb(x,y,r,col)", 7)
        x += 4
        y += 15
        for i in range(8):
            pyxel.circb(x + i * 8, y, i, i + 8)

    def test_blt(self, x, y):
        pyxel.text(x, y, "blt(x,y,img,u,v,\n    w,h,[colkey])", 7)
        y += 15
        offset = pyxel.sin(pyxel.frame_count * 5.73) * 2
        pyxel.blt(x, y, 0, 0, 0, 16, 16)
        pyxel.blt(x + offset + 19, y, 0, 0, 0, 16, 16, 13)
        pyxel.blt(x + 38, y, 0, 0, 0, -16, 16, 13)
        pyxel.blt(x + 57, y, 0, 0, 0, 16, -16, 13)
        pyxel.blt(x + 76, y, 0, 0, 0, -16, -16, 13)

    def test_bltm(self, x, y):
        pyxel.text(x, y, "bltm(x,y,tm,u,v,\n     w,h,[colkey])", 7)
        y += 15
        pyxel.bltm(x, y, 0, 0, 0, 88, 16, 2)

    def test_text(self, x, y):
        pyxel.text(x, y, "text(x,y,s,col)", 7)
        x += 4
        y += 8
        s = (
            f"Elapsed frame count is {pyxel.frame_count}\n"
            f"Current mouse position is ({pyxel.mouse_x},{pyxel.mouse_y})"
        )
        pyxel.text(x + 1, y, s, 1)
        pyxel.text(x, y, s, 9)


App()



import pyxel

from .settings import EDITOR_IMAGE, PIANO_KEYBOARD_PLAY_COLOR, PIANO_KEYBOARD_REST_COLOR
from .widgets import Widget

key_table = [
    pyxel.KEY_Z,
    pyxel.KEY_S,
    pyxel.KEY_X,
    pyxel.KEY_D,
    pyxel.KEY_C,
    pyxel.KEY_V,
    pyxel.KEY_G,
    pyxel.KEY_B,
    pyxel.KEY_H,
    pyxel.KEY_N,
    pyxel.KEY_J,
    pyxel.KEY_M,
    pyxel.KEY_Q,
    pyxel.KEY_2,
    pyxel.KEY_W,
    pyxel.KEY_3,
    pyxel.KEY_E,
    pyxel.KEY_R,
    pyxel.KEY_5,
    pyxel.KEY_T,
    pyxel.KEY_6,
    pyxel.KEY_Y,
    pyxel.KEY_7,
    pyxel.KEY_U,
]


class PianoKeyboard(Widget):
    """
    Variables:
        note_var
        octave_var
        is_playing_var
        help_message_var
    """

    def __init__(self, parent):
        super().__init__(parent, 17, 25, 12, 123)
        self._preview_sound = pyxel.Sound()
        self._preview_sound.set("g2", "p", "3", "n", 30)
        self._preview_tone = 0
        self._mouse_note = None
        self.field_cursor = parent.field_cursor
        self.get_field = parent.get_field
        self.copy_var("octave_var", parent)
        self.copy_var("is_playing_var", parent)
        self.copy_var("help_message_var", parent)

        # Initialize note_var
        self.new_var("note_var", None)

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_up", self.__on_mouse_up)
        self.add_event_listener("mouse_drag", self.__on_mouse_drag)
        self.add_event_listener("mouse_hover", self.__on_mouse_hover)
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

    def _screen_to_note(self, x, y):
        x -= self.x
        y -= self.y
        octave = (4 - y // 24) * 12
        y %= 24
        if octave > 59:
            return 59
        if octave < 0:
            return -1
        if x <= 6:
            if 2 <= y <= 4:
                return octave + 10
            elif 6 <= y <= 8:
                return octave + 8
            elif 10 <= y <= 12:
                return octave + 6
            elif 16 <= y <= 18:
                return octave + 3
            elif 20 <= y <= 22:
                return octave + 1
        if y <= 2:
            return octave + 11
        elif y <= 6:
            return octave + 9
        elif y <= 10:
            return octave + 7
        elif y <= 13:
            return octave + 5
        elif y <= 16:
            return octave + 4
        elif y <= 20:
            return octave + 2
        else:
            return octave

    def __on_mouse_down(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT:
            return
        if self.field_cursor.y > 0:
            self.field_cursor.move_to(self.field_cursor.x, 0, False)
        self._mouse_note = self._screen_to_note(x, y)

    def __on_mouse_up(self, key, x, y):
        self._mouse_note = None

    def __on_mouse_drag(self, key, x, y, dx, dy):
        self.__on_mouse_down(key, x, y)

    def __on_mouse_hover(self, x, y):
        self.help_message_var = "NOTE:Z/S/X..Q/2/W..A+ENTER TONE:1"

    def __on_update(self):
        if (
            self.field_cursor.y > 0
            or self.is_playing_var
            or pyxel.btn(pyxel.KEY_SHIFT)
            or pyxel.btn(pyxel.KEY_CTRL)
            or pyxel.btn(pyxel.KEY_ALT)
            or pyxel.btn(pyxel.KEY_GUI)
        ):
            return
        if pyxel.btnp(pyxel.KEY_1):
            self._preview_tone = (self._preview_tone + 1) % 4
        self.note_var = self._mouse_note
        for i, key in enumerate(key_table):
            if pyxel.btn(key):
                self.note_var = self.octave_var * 12 + i
                break
        if pyxel.btn(pyxel.KEY_A):
            self.note_var = -1
        if self.note_var is not None:
            self._preview_sound.notes[0] = self.note_var
            self._preview_sound.tones[0] = self._preview_tone
            pyxel.play(1, self._preview_sound)
        else:
            pyxel.stop(1)

    def __on_draw(self):
        pyxel.blt(
            self.x,
            self.y,
            EDITOR_IMAGE,
            208,
            0,
            12,
            123,
        )
        notes = self.get_field(0)
        if (
            self.is_playing_var
            and not notes
            or not self.is_playing_var
            and self.note_var is None
        ):
            return
        note = notes[pyxel.play_pos(0)[1]] if self.is_playing_var else self.note_var
        key = note % 12
        x = self.x
        y = self.y + (59 - note) * 2
        if note == -1:
            pyxel.rect(x, y + 1, 12, 2, PIANO_KEYBOARD_REST_COLOR)
        elif key == 0 or key == 5:
            pyxel.rect(x, y + 1, 7, 1, PIANO_KEYBOARD_PLAY_COLOR)
            pyxel.rect(x + 7, y, 5, 2, PIANO_KEYBOARD_PLAY_COLOR)
        elif key == 4 or key == 11:
            pyxel.rect(x, y + 1, 7, 1, PIANO_KEYBOARD_PLAY_COLOR)
            pyxel.rect(x + 7, y + 1, 5, 2, PIANO_KEYBOARD_PLAY_COLOR)
        elif key == 2 or key == 7 or key == 9:
            pyxel.rect(x, y + 1, 7, 1, PIANO_KEYBOARD_PLAY_COLOR)
            pyxel.rect(x + 7, y, 5, 3, PIANO_KEYBOARD_PLAY_COLOR)
        else:
            pyxel.rect(x, y + 1, 6, 1, PIANO_KEYBOARD_PLAY_COLOR)



import os

import pyxel

from .image_editor import ImageEditor
from .music_editor import MusicEditor
from .settings import APP_HEIGHT, APP_WIDTH, EDITOR_IMAGE, HELP_MESSAGE_COLOR
from .sound_editor import SoundEditor
from .tilemap_editor import TilemapEditor
from .widgets import ImageButton, RadioButton, Widget
from .widgets.settings import (
    WIDGET_BACKGROUND_COLOR,
    WIDGET_HOLD_TIME,
    WIDGET_PANEL_COLOR,
    WIDGET_REPEAT_TIME,
    WIDGET_SHADOW_COLOR,
)


class App(Widget):
    """
    Variables:
        editor_no_var
        help_message_var
    """

    def __init__(self, resource_file):
        # Get absolute path of resource file before initializing Pyxel
        resource_file = os.path.abspath(resource_file)

        # Initialize Pyxel
        pyxel.init(APP_WIDTH, APP_HEIGHT)
        pyxel.mouse(True)
        self._set_title(resource_file)
        if os.path.exists(resource_file):
            pyxel.load(resource_file)

        # Start initializing application
        super().__init__(None, 0, 0, pyxel.width, pyxel.height)
        self._resource_file = resource_file

        # Initialize help_message_var
        self.new_var("help_message_var", "")

        # Initialize editor button
        self._editor_button = RadioButton(
            self,
            1,
            1,
            img=EDITOR_IMAGE,
            u=0,
            v=0,
            num_buttons=4,
            value=0,
        )
        self._editor_button.add_event_listener("change", self.__on_editor_button_change)
        self._editor_button.add_event_listener(
            "mouse_hover", self.__on_editor_button_mouse_hover
        )
        self.copy_var("editor_no_var", self._editor_button, "value_var")

        # Initialize undo button
        self._undo_button = ImageButton(
            self,
            48,
            1,
            img=EDITOR_IMAGE,
            u=36,
            v=0,
        )
        self._undo_button.add_event_listener("press", self.__on_undo_button_press)
        self._undo_button.add_event_listener(
            "mouse_hover", self.__on_undo_button_mouse_hover
        )

        # Initialize redo button
        self._redo_button = ImageButton(
            self,
            57,
            1,
            img=EDITOR_IMAGE,
            u=45,
            v=0,
        )
        self._redo_button.add_event_listener("press", self.__on_redo_button_press)
        self._redo_button.add_event_listener(
            "mouse_hover", self.__on_redo_button_mouse_hover
        )

        # Initialize save button
        self._save_button = ImageButton(
            self,
            75,
            1,
            img=EDITOR_IMAGE,
            u=54,
            v=0,
        )
        self._save_button.add_event_listener("press", self.__on_save_button_press)
        self._save_button.add_event_listener(
            "mouse_hover", self.__on_save_button_mouse_hover
        )

        # Initialize editors
        self._editors = [
            ImageEditor(self),
            TilemapEditor(self),
            SoundEditor(self),
            MusicEditor(self),
        ]
        self.__on_editor_button_change(self.editor_no_var)

        # Set event listeners
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

        # Start application
        pyxel.run(self.update_all, self.draw_all)

    @property
    def _editor(self):
        return self._editors[self.editor_no_var]

    @staticmethod
    def _set_title(filename):
        pyxel.title(f"Pyxel Editor - {filename}")

    def __on_editor_button_change(self, value):
        for i, editor in enumerate(self._editors):
            editor.is_visible_var = i == value

    def __on_editor_button_mouse_hover(self, x, y):
        self.help_message_var = "EDITOR:ALT+LEFT/RIGHT"

    def __on_undo_button_press(self):
        self._editor.undo()

    def __on_undo_button_mouse_hover(self, x, y):
        self.help_message_var = "UNDO:CTRL+Z"

    def __on_redo_button_press(self):
        self._editor.redo()

    def __on_redo_button_mouse_hover(self, x, y):
        self.help_message_var = "REDO:CTRL+Y"

    def __on_save_button_press(self):
        pyxel.save(self._resource_file)

    def __on_save_button_mouse_hover(self, x, y):
        self.help_message_var = "SAVE:CTRL+S"

    def __on_update(self):
        if pyxel.drop_files:
            drop_file = pyxel.drop_files[-1]
            file_ext = os.path.splitext(drop_file)[1]
            if file_ext == pyxel.RESOURCE_FILE_EXTENSION:
                pyxel.stop()
                if pyxel.btn(pyxel.KEY_CTRL) or pyxel.btn(pyxel.KEY_GUI):
                    self._editor.reset_history()
                    pyxel.load(
                        pyxel._drop_file,
                        image=(self.editor_no_var == 0),
                        tilemap=(self.editor_no_var == 1),
                        sound=(self.editor_no_var == 2),
                        music=(self.editor_no_var == 3),
                    )
                else:
                    for editor in self._editors:
                        editor.reset_history()
                    pyxel.load(drop_file)
                    self._set_title(drop_file)
            else:
                self._editor.trigger_event("drop", drop_file)

        if pyxel.btn(pyxel.KEY_ALT):
            # Alt+Left: Switch editor
            if pyxel.btnp(pyxel.KEY_LEFT):
                self.editor_no_var = (self.editor_no_var - 1) % len(self._editors)

            # Alt+Right: Switch editor
            elif pyxel.btnp(pyxel.KEY_RIGHT):
                self.editor_no_var = (self.editor_no_var + 1) % len(self._editors)

        self._undo_button.is_enabled_var = self._editor.can_undo
        self._redo_button.is_enabled_var = self._editor.can_redo

        if pyxel.btn(pyxel.KEY_CTRL) or pyxel.btn(pyxel.KEY_GUI):
            # Ctrl+S: Save
            if pyxel.btnp(pyxel.KEY_S):
                self._save_button.is_pressed_var = True

            # Ctrl+Z: Undo
            if self._editor.can_undo and pyxel.btnp(
                pyxel.KEY_Z, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME
            ):
                self._undo_button.is_pressed_var = True

            # Ctrl+Y: Redo
            elif self._editor.can_redo and pyxel.btnp(
                pyxel.KEY_Y, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME
            ):
                self._redo_button.is_pressed_var = True

    def __on_draw(self):
        pyxel.cls(WIDGET_BACKGROUND_COLOR)
        pyxel.rect(0, 0, 240, 9, WIDGET_PANEL_COLOR)
        pyxel.line(0, 9, 239, 9, WIDGET_SHADOW_COLOR)
        pyxel.text(93, 2, self.help_message_var, HELP_MESSAGE_COLOR)
        self.help_message_var = ""



class WidgetVar:
    """
    Events:
        get (value) -> value
        set (value) -> value
        change (value)
    """

    def __init__(self, value):
        self._value = value
        self._event_listeners = {"get": [], "set": [], "change": []}

    def get(self):
        value = self._value
        for listener in self._event_listeners["get"]:
            value = listener(value)
        return value

    def set(self, value):
        for listener in self._event_listeners["set"]:
            value = listener(value)
        if self._value == value:
            return
        self._value = value
        for listener in self._event_listeners["change"]:
            listener(value)

    def add_event_listener(self, event, listener):
        self._event_listeners[event].append(listener)

    def remove_event_listener(self, event, listener):
        self._event_listeners[event].remove(listener)



import pyxel

from .settings import BUTTON_ENABLED_COLOR
from .toggle_button import ToggleButton


class ImageToggleButton(ToggleButton):
    """
    Variables:
        is_checked_var

    Events:
        checked
        unchecked
    """

    def __init__(self, parent, x, y, *, img, u, v, is_checked, **kwargs):
        super().__init__(parent, x, y, 7, 7, is_checked=is_checked, **kwargs)
        self._img = img
        self._u = u
        self._v = v

        # Set event listeners
        self.add_event_listener("draw", self.__on_draw)

    def __on_draw(self):
        pyxel.pal(BUTTON_ENABLED_COLOR, self.button_color)
        pyxel.blt(
            self.x, self.y, self._img, self._u, self._v, self.width, self.height, 0
        )
        pyxel.pal()



import pyxel

from .settings import BUTTON_ENABLED_COLOR, BUTTON_PRESSED_COLOR
from .widget import Widget


class RadioButton(Widget):
    """
    Variables:
        value_var

    Events:
        change (value)
    """

    def __init__(self, parent, x, y, *, img, u, v, num_buttons, value, **kwargs):
        super().__init__(parent, x, y, num_buttons * 9 - 2, 7, **kwargs)
        self._img = img
        self._u = u
        self._v = v
        self._num_buttons = num_buttons

        # Initialize value_var
        self.new_var("value_var", value)
        self.add_var_event_listener("value_var", "change", self.__on_value_change)

        # Initialize event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_drag", self.__on_mouse_drag)
        self.add_event_listener("draw", self.__on_draw)

    def check_value(self, x, y):
        x -= self.x
        y -= self.y
        index = min(max(x // 9, 0), self._num_buttons - 1)
        x1 = index * 9
        y1 = 0
        x2 = x1 + 6
        y2 = y1 + 6
        return index if x1 <= x <= x2 and y1 <= y <= y2 else None

    def __on_value_change(self, value):
        self.trigger_event("change", value)

    def __on_mouse_down(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT:
            return
        value = self.check_value(x, y)
        if value is not None:
            self.value_var = value

    def __on_mouse_drag(self, key, x, y, dx, dy):
        self.__on_mouse_down(key, x, y)

    def __on_draw(self):
        pyxel.blt(
            self.x,
            self.y,
            self._img,
            self._u,
            self._v,
            self.width,
            self.height,
        )
        pyxel.pal(BUTTON_ENABLED_COLOR, BUTTON_PRESSED_COLOR)
        pyxel.blt(
            self.x + self.value_var * 9,
            self.y,
            self._img,
            self._u + self.value_var * 9,
            self._v,
            7,
            7,
        )
        pyxel.pal()



# flake8: noqa
from typing import Callable, List, Optional, Tuple, Union

# Constants
PYXEL_VERSION: str
PYXEL_WORKING_DIR: str

APP_FILE_EXTENSION: str
APP_STARTUP_SCRIPT_FILE: str
RESOURCE_FILE_EXTENSION: str
RESOURCE_ARCHIVE_DIRNAME: str

NUM_COLORS: int
NUM_IMAGES: int
IMAGE_SIZE: int
NUM_TILEMAPS: int
TILEMAP_SIZE: int
TILE_SIZE: int

COLOR_BLACK: int
COLOR_NAVY: int
COLOR_PURPLE: int
COLOR_GREEN: int
COLOR_BROWN: int
COLOR_DARK_BLUE: int
COLOR_LIGHT_BLUE: int
COLOR_WHITE: int
COLOR_RED: int
COLOR_ORANGE: int
COLOR_YELLOW: int
COLOR_LIME: int
COLOR_CYAN: int
COLOR_GRAY: int
COLOR_PINK: int
COLOR_PEACH: int

FONT_WIDTH: int
FONT_HEIGHT: int

NUM_CHANNELS: int
NUM_SOUNDS: int
NUM_MUSICS: int

TONE_TRIANGLE: int
TONE_SQUARE: int
TONE_PULSE: int
TONE_NOISE: int

EFFECT_NONE: int
EFFECT_SLIDE: int
EFFECT_VIBRATO: int
EFFECT_FADEOUT: int

# Keys
KEY_BACKSPACE: int
KEY_TAB: int
KEY_RETURN: int
KEY_ESCAPE: int
KEY_SPACE: int
KEY_EXCLAIM: int
KEY_QUOTEDBL: int
KEY_HASH: int
KEY_DOLLAR: int
KEY_PERCENT: int
KEY_AMPERSAND: int
KEY_QUOTE: int
KEY_LEFTPAREN: int
KEY_RIGHTPAREN: int
KEY_ASTERISK: int
KEY_PLUS: int
KEY_COMMA: int
KEY_MINUS: int
KEY_PERIOD: int
KEY_SLASH: int
KEY_0: int
KEY_1: int
KEY_2: int
KEY_3: int
KEY_4: int
KEY_5: int
KEY_6: int
KEY_7: int
KEY_8: int
KEY_9: int
KEY_COLON: int
KEY_SEMICOLON: int
KEY_LESS: int
KEY_EQUALS: int
KEY_GREATER: int
KEY_QUESTION: int
KEY_AT: int
KEY_LEFTBRACKET: int
KEY_BACKSLASH: int
KEY_RIGHTBRACKET: int
KEY_CARET: int
KEY_UNDERSCORE: int
KEY_BACKQUOTE: int
KEY_A: int
KEY_B: int
KEY_C: int
KEY_D: int
KEY_E: int
KEY_F: int
KEY_G: int
KEY_H: int
KEY_I: int
KEY_J: int
KEY_K: int
KEY_L: int
KEY_M: int
KEY_N: int
KEY_O: int
KEY_P: int
KEY_Q: int
KEY_R: int
KEY_S: int
KEY_T: int
KEY_U: int
KEY_V: int
KEY_W: int
KEY_X: int
KEY_Y: int
KEY_Z: int
KEY_DELETE: int
KEY_CAPSLOCK: int
KEY_F1: int
KEY_F2: int
KEY_F3: int
KEY_F4: int
KEY_F5: int
KEY_F6: int
KEY_F7: int
KEY_F8: int
KEY_F9: int
KEY_F10: int
KEY_F11: int
KEY_F12: int
KEY_PRINTSCREEN: int
KEY_SCROLLLOCK: int
KEY_PAUSE: int
KEY_INSERT: int
KEY_HOME: int
KEY_PAGEUP: int
KEY_END: int
KEY_PAGEDOWN: int
KEY_RIGHT: int
KEY_LEFT: int
KEY_DOWN: int
KEY_UP: int
KEY_NUMLOCKCLEAR: int
KEY_KP_DIVIDE: int
KEY_KP_MULTIPLY: int
KEY_KP_MINUS: int
KEY_KP_PLUS: int
KEY_KP_ENTER: int
KEY_KP_1: int
KEY_KP_2: int
KEY_KP_3: int
KEY_KP_4: int
KEY_KP_5: int
KEY_KP_6: int
KEY_KP_7: int
KEY_KP_8: int
KEY_KP_9: int
KEY_KP_0: int
KEY_KP_PERIOD: int
KEY_APPLICATION: int
KEY_POWER: int
KEY_KP_EQUALS: int
KEY_F13: int
KEY_F14: int
KEY_F15: int
KEY_F16: int
KEY_F17: int
KEY_F18: int
KEY_F19: int
KEY_F20: int
KEY_F21: int
KEY_F22: int
KEY_F23: int
KEY_F24: int
KEY_EXECUTE: int
KEY_HELP: int
KEY_MENU: int
KEY_SELECT: int
KEY_STOP: int
KEY_AGAIN: int
KEY_UNDO: int
KEY_CUT: int
KEY_COPY: int
KEY_PASTE: int
KEY_FIND: int
KEY_MUTE: int
KEY_VOLUMEUP: int
KEY_VOLUMEDOWN: int
KEY_KP_COMMA: int
KEY_KP_EQUALSAS400: int
KEY_ALTERASE: int
KEY_SYSREQ: int
KEY_CANCEL: int
KEY_CLEAR: int
KEY_PRIOR: int
KEY_RETURN2: int
KEY_SEPARATOR: int
KEY_OUT: int
KEY_OPER: int
KEY_CLEARAGAIN: int
KEY_CRSEL: int
KEY_EXSEL: int
KEY_KP_00: int
KEY_KP_000: int
KEY_THOUSANDSSEPARATOR: int
KEY_DECIMALSEPARATOR: int
KEY_CURRENCYUNIT: int
KEY_CURRENCYSUBUNIT: int
KEY_KP_LEFTPAREN: int
KEY_KP_RIGHTPAREN: int
KEY_KP_LEFTBRACE: int
KEY_KP_RIGHTBRACE: int
KEY_KP_TAB: int
KEY_KP_BACKSPACE: int
KEY_KP_A: int
KEY_KP_B: int
KEY_KP_C: int
KEY_KP_D: int
KEY_KP_E: int
KEY_KP_F: int
KEY_KP_XOR: int
KEY_KP_POWER: int
KEY_KP_PERCENT: int
KEY_KP_LESS: int
KEY_KP_GREATER: int
KEY_KP_AMPERSAND: int
KEY_KP_DBLAMPERSAND: int
KEY_KP_VERTICALBAR: int
KEY_KP_DBLVERTICALBAR: int
KEY_KP_COLON: int
KEY_KP_HASH: int
KEY_KP_SPACE: int
KEY_KP_AT: int
KEY_KP_EXCLAM: int
KEY_KP_MEMSTORE: int
KEY_KP_MEMRECALL: int
KEY_KP_MEMCLEAR: int
KEY_KP_MEMADD: int
KEY_KP_MEMSUBTRACT: int
KEY_KP_MEMMULTIPLY: int
KEY_KP_MEMDIVIDE: int
KEY_KP_PLUSMINUS: int
KEY_KP_CLEAR: int
KEY_KP_CLEARENTRY: int
KEY_KP_BINARY: int
KEY_KP_OCTAL: int
KEY_KP_DECIMAL: int
KEY_KP_HEXADECIMAL: int
KEY_LCTRL: int
KEY_LSHIFT: int
KEY_LALT: int
KEY_LGUI: int
KEY_RCTRL: int
KEY_RSHIFT: int
KEY_RALT: int
KEY_RGUI: int
KEY_MODE: int
KEY_AUDIONEXT: int
KEY_AUDIOPREV: int
KEY_AUDIOSTOP: int
KEY_AUDIOPLAY: int
KEY_AUDIOMUTE: int
KEY_MEDIASELECT: int
KEY_WWW: int
KEY_MAIL: int
KEY_CALCULATOR: int
KEY_COMPUTER: int
KEY_AC_SEARCH: int
KEY_AC_HOME: int
KEY_AC_BACK: int
KEY_AC_FORWARD: int
KEY_AC_STOP: int
KEY_AC_REFRESH: int
KEY_AC_BOOKMARKS: int
KEY_BRIGHTNESSDOWN: int
KEY_BRIGHTNESSUP: int
KEY_DISPLAYSWITCH: int
KEY_KBDILLUMTOGGLE: int
KEY_KBDILLUMDOWN: int
KEY_KBDILLUMUP: int
KEY_EJECT: int
KEY_SLEEP: int
KEY_NONE: int
KEY_SHIFT: int
KEY_CTRL: int
KEY_ALT: int
KEY_GUI: int

MOUSE_POS_X: int
MOUSE_POS_Y: int
MOUSE_WHEEL_X: int
MOUSE_WHEEL_Y: int
MOUSE_BUTTON_LEFT: int
MOUSE_BUTTON_MIDDLE: int
MOUSE_BUTTON_RIGHT: int
MOUSE_BUTTON_X1: int
MOUSE_BUTTON_X2: int
MOUSE_BUTTON_UNKNOWN: int

GAMEPAD1_AXIS_LEFTX: int
GAMEPAD1_AXIS_LEFTY: int
GAMEPAD1_AXIS_RIGHTX: int
GAMEPAD1_AXIS_RIGHTY: int
GAMEPAD1_AXIS_TRIGGERLEFT: int
GAMEPAD1_AXIS_TRIGGERRIGHT: int
GAMEPAD1_BUTTON_A: int
GAMEPAD1_BUTTON_B: int
GAMEPAD1_BUTTON_X: int
GAMEPAD1_BUTTON_Y: int
GAMEPAD1_BUTTON_BACK: int
GAMEPAD1_BUTTON_GUIDE: int
GAMEPAD1_BUTTON_START: int
GAMEPAD1_BUTTON_LEFTSTICK: int
GAMEPAD1_BUTTON_RIGHTSTICK: int
GAMEPAD1_BUTTON_LEFTSHOULDER: int
GAMEPAD1_BUTTON_RIGHTSHOULDER: int
GAMEPAD1_BUTTON_DPAD_UP: int
GAMEPAD1_BUTTON_DPAD_DOWN: int
GAMEPAD1_BUTTON_DPAD_LEFT: int
GAMEPAD1_BUTTON_DPAD_RIGHT: int

GAMEPAD2_AXIS_LEFTX: int
GAMEPAD2_AXIS_LEFTY: int
GAMEPAD2_AXIS_RIGHTX: int
GAMEPAD2_AXIS_RIGHTY: int
GAMEPAD2_AXIS_TRIGGERLEFT: int
GAMEPAD2_AXIS_TRIGGERRIGHT: int
GAMEPAD2_BUTTON_A: int
GAMEPAD2_BUTTON_B: int
GAMEPAD2_BUTTON_X: int
GAMEPAD2_BUTTON_Y: int
GAMEPAD2_BUTTON_BACK: int
GAMEPAD2_BUTTON_GUIDE: int
GAMEPAD2_BUTTON_START: int
GAMEPAD2_BUTTON_LEFTSTICK: int
GAMEPAD2_BUTTON_RIGHTSTICK: int
GAMEPAD2_BUTTON_LEFTSHOULDER: int
GAMEPAD2_BUTTON_RIGHTSHOULDER: int
GAMEPAD2_BUTTON_DPAD_UP: int
GAMEPAD2_BUTTON_DPAD_DOWN: int
GAMEPAD2_BUTTON_DPAD_LEFT: int
GAMEPAD2_BUTTON_DPAD_RIGHT: int

# System
width: int
height: int
frame_count: int
is_fullscreen: bool

def init(
    width: int,
    height: int,
    *,
    title: Optional[str] = None,
    fps: Optional[int] = None,
    quit_key: Optional[int] = None,
    display_scale: Optional[int] = None,
    capture_scale: Optional[int] = None,
    capture_sec: Optional[int] = None,
) -> None: ...
def title(title: str) -> None: ...
def icon(data: List[str], scale: int) -> None: ...
def fullscreen(full: bool) -> None: ...
def run(update: Callable[[], None], draw: Callable[[], None]) -> None: ...
def show() -> None: ...
def quit() -> None: ...
def process_exists(pid: int) -> bool: ...

# Resource
def load(
    filename: str,
    *,
    image: Optional[bool] = None,
    tilemap: Optional[bool] = None,
    sound: Optional[bool] = None,
    music: Optional[bool] = None,
) -> None: ...
def save(
    filename: str,
    *,
    image: Optional[bool] = None,
    tilemap: Optional[bool] = None,
    sound: Optional[bool] = None,
    music: Optional[bool] = None,
) -> None: ...
def screenshot(scale: Optional[int] = None) -> None: ...
def reset_capture() -> None: ...
def screencast(scale: Optional[int] = None) -> None: ...

# Input
mouse_x: int
mouse_y: int
mouse_wheel: int
input_keys: List[int]
input_text: str
drop_files: List[str]

def btn(key: int) -> bool: ...
def btnp(
    key: int, *, hold: Optional[int] = None, repeat: Optional[int] = None
) -> bool: ...
def btnr(key: int) -> bool: ...
def btnv(key: int) -> int: ...
def mouse(visible: bool) -> None: ...
def set_btn(key: int, state: bool) -> None: ...
def set_btnv(key: int, val: float) -> None: ...
def set_mouse_pos(x: float, y: float) -> None: ...

# Graphics
class Image: ...
class Tilemap: ...

colors: List[int]
screen: Image
cursor: Image
font: Image

def image(img: int) -> Image: ...
def tilemap(tm: int) -> Tilemap: ...
def clip(
    x: Optional[float] = None,
    y: Optional[float] = None,
    w: Optional[float] = None,
    h: Optional[float] = None,
) -> None: ...
def camera(
    x: Optional[float] = None,
    y: Optional[float] = None,
) -> None: ...
def pal(col1: Optional[int] = None, col2: Optional[int] = None) -> None: ...
def cls(col: int) -> None: ...
def pget(x: float, y: float) -> int: ...
def pset(x: float, y: float, col: int) -> None: ...
def line(x1: float, y1: float, x2: float, y2: float, col: int) -> None: ...
def rect(x: float, y: float, w: float, h: float, col: int) -> None: ...
def rectb(x: float, y: float, w: float, h: float, col: int) -> None: ...
def circ(x: float, y: float, r: float, col: int) -> None: ...
def circb(x: float, y: float, r: float, col: int) -> None: ...
def elli(self, x: float, y: float, w: float, h: float, col: int) -> None: ...
def ellib(self, x: float, y: float, w: float, h: float, col: int) -> None: ...
def tri(
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    col: int,
) -> None: ...
def trib(
    x1: float,
    y1: float,
    x2: float,
    y2: float,
    x3: float,
    y3: float,
    col: int,
) -> None: ...
def fill(x: float, y: float, col: int) -> None: ...
def blt(
    x: float,
    y: float,
    img: Union[int, Image],
    u: float,
    v: float,
    w: float,
    h: float,
    colkey: Optional[int] = None,
) -> None: ...
def bltm(
    x: float,
    y: float,
    tm: Union[int, Tilemap],
    u: float,
    v: float,
    w: float,
    h: float,
    colkey: Optional[int] = None,
) -> None: ...
def text(x: float, y: float, s: str, col: int) -> None: ...

# Audio
class Channel: ...
class Sound: ...
class Music: ...

def channel(ch: int) -> Channel: ...
def sound(snd: int) -> Sound: ...
def music(msc: int) -> Music: ...
def play_pos(ch: int) -> Optional[Tuple[int, int]]: ...
def play(
    ch: int,
    snd: Union[int, List[int], Sound, List[Sound]],
    *,
    tick: Optional[int] = None,
    loop: Optional[bool] = None,
) -> None: ...
def playm(
    msc: int, *, tick: Optional[int] = None, loop: Optional[bool] = None
) -> None: ...
def stop(ch: Optional[int] = None) -> None: ...

# Math
def ceil(x: float) -> int: ...
def floor(x: float) -> int: ...
def sgn(x: float) -> float: ...
def sqrt(x: float) -> float: ...
def sin(deg: float) -> float: ...
def cos(deg: float) -> float: ...
def atan2(y: float, x: float) -> float: ...
def rseed(seed: int) -> None: ...
def rndi(a: int, b: int) -> int: ...
def rndf(a: float, b: float) -> int: ...
def nseed(seed: int) -> None: ...
def noise(x: float, y: Optional[float] = None, z: Optional[float] = None) -> float: ...

# Image class
class Image:
    width: int
    height: int
    def __init__(self, width: int, height: int) -> None: ...
    @staticmethod
    def from_image(filename: str) -> Image: ...
    def set(self, x: int, y: int, data: List[str]) -> None: ...
    def load(self, x: int, y: int, filename: str) -> None: ...
    def save(self, filename: str, scale: int) -> None: ...
    def clip(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        w: Optional[float] = None,
        h: Optional[float] = None,
    ) -> None: ...
    def camera(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
    ) -> None: ...
    def pal(self, col1: Optional[int] = None, col2: Optional[int] = None) -> None: ...
    def cls(self, col: int) -> None: ...
    def pget(self, x: float, y: float) -> int: ...
    def pset(self, x: float, y: float, col: int) -> None: ...
    def line(self, x1: float, y1: float, x2: float, y2: float, col: int) -> None: ...
    def rect(self, x: float, y: float, w: float, h: float, col: int) -> None: ...
    def rectb(self, x: float, y: float, w: float, h: float, col: int) -> None: ...
    def circ(self, x: float, y: float, r: float, col: int) -> None: ...
    def circb(self, x: float, y: float, r: float, col: int) -> None: ...
    def elli(self, x: float, y: float, w: float, h: float, col: int) -> None: ...
    def ellib(self, x: float, y: float, w: float, h: float, col: int) -> None: ...
    def tri(
        self, x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, col: int
    ) -> None: ...
    def trib(
        self, x1: float, y1: float, x2: float, y2: float, x3: float, y3: float, col: int
    ) -> None: ...
    def fill(self, x: float, y: float, col: int) -> None: ...
    def blt(
        self,
        x: float,
        y: float,
        img: Union[int, Image],
        u: float,
        v: float,
        w: float,
        h: float,
        colkey: Optional[int] = None,
    ) -> None: ...
    def bltm(
        self,
        x: float,
        y: float,
        tm: Union[int, Tilemap],
        u: float,
        v: float,
        w: float,
        h: float,
        colkey: Optional[int] = None,
    ) -> None: ...
    def text(self, x: float, y: float, s: str, col: int) -> None: ...

# Tilemap class
class Tilemap:
    width: int
    height: int
    image: Image
    refimg: Optional[int]
    def __init__(self, width: int, height: int, img: Union[int, Image]) -> None: ...
    def set(self, x: int, y: int, data: List[str]) -> None: ...
    def clip(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        w: Optional[float] = None,
        h: Optional[float] = None,
    ) -> None: ...
    def camera(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
    ) -> None: ...
    def cls(self, tile: Tuple[int, int]) -> None: ...
    def pget(self, x: float, y: float) -> Tuple[int, int]: ...
    def pset(self, x: float, y: float, tile: Tuple[int, int]) -> None: ...
    def line(
        self, x1: float, y1: float, x2: float, y2: float, tile: Tuple[int, int]
    ) -> None: ...
    def rect(
        self, x: float, y: float, w: float, h: float, tile: Tuple[int, int]
    ) -> None: ...
    def rectb(
        self, x: float, y: float, w: float, h: float, tile: Tuple[int, int]
    ) -> None: ...
    def circ(self, x: float, y: float, r: float, tile: Tuple[int, int]) -> None: ...
    def circb(self, x: float, y: float, r: float, tile: Tuple[int, int]) -> None: ...
    def elli(
        self, x: float, y: float, w: float, h: float, tile: Tuple[int, int]
    ) -> None: ...
    def ellib(
        self, x: float, y: float, w: float, h: float, tile: Tuple[int, int]
    ) -> None: ...
    def tri(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        x3: float,
        y3: float,
        tile: Tuple[int, int],
    ) -> None: ...
    def trib(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        x3: float,
        y3: float,
        tile: Tuple[int, int],
    ) -> None: ...
    def fill(self, x: float, y: float, tile: Tuple[int, int]) -> None: ...
    def blt(
        self,
        x: float,
        y: float,
        tm: Union[int, Tilemap],
        u: float,
        v: float,
        w: float,
        h: float,
        tilekey: Optional[Tuple[int, int]] = None,
    ) -> None: ...

# Channel class
class Channel:
    gain: int
    def play_pos(self) -> Optional[Tuple[int, int]]: ...
    def play(
        self,
        snd: Union[int, List[int], Sound, List[Sound]],
        *,
        tick: Optional[int] = None,
        loop: Optional[bool] = None,
    ) -> None: ...
    def stop(self) -> None: ...

# Sound class
class Sound:
    notes: List[int]
    tones: List[int]
    volumes: List[int]
    effects: List[int]
    speed: int
    def __init__(self) -> None: ...
    def set(
        self,
        notes: str,
        tones: str,
        volumes: str,
        effects: str,
        speed: int,
    ) -> None: ...
    def set_notes(self, notes: str) -> None: ...
    def set_tones(self, tones: str) -> None: ...
    def set_volumes(self, volumes: str) -> None: ...
    def set_effects(self, effects: str) -> None: ...

# Music class
class Music:
    snds_list: List[List[int]]
    def __init__(self) -> None: ...
    def set(
        self,
        snds0: List[int],
        snds1: List[int],
        snds2: List[int],
        snds3: List[int],
    ) -> None: ...



import pyxel

from .widgets.settings import WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME


class FieldCursor:
    def __init__(
        self,
        *,
        max_field_length,
        field_wrap_length,
        get_field,
        add_pre_history,
        add_post_history,
        cross_filed_copying,
    ):
        self._max_field_length = max_field_length
        self._field_wrap_length = (
            field_wrap_length
            if field_wrap_length < max_field_length
            else max_field_length + 1
        )
        self._get_field = get_field
        self._add_pre_history = add_pre_history
        self._add_post_history = add_post_history
        self._cross_field_copying = cross_filed_copying
        self._cursor_x = 0
        self._cursor_y = 0
        self._select_x = None
        self._copy_field = None

    @property
    def x(self):
        return (
            min(self._adjusted_cursor_x, self._adjusted_select_x)
            if self.is_selecting
            else self._adjusted_cursor_x
        )

    @property
    def y(self):
        return self._cursor_y

    @property
    def width(self):
        if self.is_selecting:
            width = abs(self._adjusted_cursor_x - self._adjusted_select_x) + 1
            return min(width, len(self.field) - self.x)
        else:
            return 1

    @property
    def field(self):
        return self._get_field(self._cursor_y)

    @property
    def is_selecting(self):
        return self._select_x is not None and len(self.field) > 0

    @property
    def _max_cursor_x(self):
        return max(min(len(self.field), self._max_field_length - 1), 0)

    @property
    def _max_select_x(self):
        return max(min(len(self.field) - 1, self._max_field_length - 1), 0)

    @property
    def _max_y(self):
        y = 0
        while self._get_field(y + 1) is not None:
            y += 1
        return y

    @property
    def _adjusted_cursor_x(self):
        return min(self._cursor_x, self._max_cursor_x)

    @property
    def _adjusted_select_x(self):
        return min(self._select_x, self._max_select_x)

    def move_to(self, x, y, with_select_key):
        y = max(min(y, self._max_y), 0)
        if self._cursor_y != y:
            self._cursor_x = max(min(x, self._max_cursor_x), 0)
            self._cursor_y = y
            self._select_x = None
        elif with_select_key:
            if self.is_selecting:
                self._cursor_x = max(min(x, self._max_select_x), 0)
            else:
                self._select_x = max(
                    min(self._adjusted_cursor_x, self._max_select_x), 0
                )
                self._cursor_x = max(min(x, self._max_select_x), 0)
        else:
            self._cursor_x = max(min(x, self._max_cursor_x), 0)
            self._select_x = None

    def move_left(self, with_select_key):
        if with_select_key:
            if self.is_selecting:
                self._cursor_x = max(self._adjusted_cursor_x - 1, 0)
            elif len(self.field) > 0:
                self._cursor_x = self._select_x = min(
                    self._adjusted_cursor_x, self._max_select_x
                )
        else:
            self._cursor_x = max(self._adjusted_cursor_x - 1, 0)
            self._select_x = None

    def move_right(self, with_select_key):
        if with_select_key:
            if self.is_selecting:
                self._cursor_x = min(self._adjusted_cursor_x + 1, self._max_select_x)
            elif self._adjusted_cursor_x <= self._max_select_x:
                self._cursor_x = self._select_x = min(
                    self._adjusted_cursor_x, self._max_select_x
                )
        else:
            self._cursor_x = min(self._adjusted_cursor_x + 1, self._max_cursor_x)
            self._select_x = None

    def move_up(self, with_select_key):
        if self._adjusted_cursor_x >= self._field_wrap_length:
            if not with_select_key:
                self._select_x = None
            elif not self.is_selecting:
                self._select_x = self._adjusted_cursor_x
            self._cursor_x -= self._field_wrap_length
        elif self._cursor_y > 0:
            self._cursor_y -= 1
            self._cursor_x = (
                self._field_wrap_length * (len(self.field) // self._field_wrap_length)
                + self._cursor_x % self._field_wrap_length
            )
            self._select_x = None

    def move_down(self, with_select_key):
        if (
            self._adjusted_cursor_x // self._field_wrap_length
            < len(self.field) // self._field_wrap_length
        ):
            if not with_select_key:
                self._select_x = None
            elif not self.is_selecting:
                self._select_x = self._adjusted_cursor_x
            self._cursor_x += self._field_wrap_length
        elif self._cursor_y < self._max_y:
            self._cursor_y += 1
            self._cursor_x %= self._field_wrap_length
            self._select_x = None

    def insert(self, value):
        self._add_pre_history(self.x, self.y)
        lst = self.field.to_list()
        x = self.x
        if self.is_selecting:
            lst[x : x + self.width] = []
        if not isinstance(value, list):
            value = [value]
        lst[x:x] = value
        self.field.from_list(lst[: self._max_field_length])
        self.move_to(x + len(value), self.y, False)
        self._add_post_history(self.x, self.y)

    def backspace(self):
        if not self.is_selecting and self.x == 0:
            return
        self._add_pre_history(self.x, self.y)
        lst = self.field.to_list()
        if self.is_selecting:
            x = self.x
            width = self.width
        else:
            x = self.x - 1
            width = 1
        lst[x : x + width] = []
        self.field.from_list(lst)
        self.move_to(x, self.y, False)
        self._add_post_history(self.x, self.y)

    def delete(self):
        if self.x >= len(self.field):
            return
        self._add_pre_history(self.x, self.y)
        lst = self.field.to_list()
        x = self.x
        width = self.width
        lst[x : x + width] = []
        self.field.from_list(lst)
        self.move_to(x, self.y, False)
        self._add_post_history(self.x, self.y)

    def select_all(self):
        if len(self.field) == 0:
            return
        self._cursor_x = 0
        self._select_x = len(self.field) - 1

    def copy(self):
        if not self.is_selecting:
            return
        lst = self.field.to_list()
        self._copy_field = (self.y, lst[self.x : self.x + self.width])

    def cut(self):
        if not self.is_selecting:
            return
        self.copy()
        self.delete()

    def paste(self):
        if self._copy_field is None:
            return
        (y, field) = self._copy_field
        if not self._cross_field_copying and self.y != y:
            return
        self.insert(field)

    def process_input(self):
        if pyxel.btn(pyxel.KEY_ALT):
            return
        if pyxel.btn(pyxel.KEY_CTRL) or pyxel.btn(pyxel.KEY_GUI):
            # Ctrl+A: Select all
            if pyxel.btnp(pyxel.KEY_A):
                self.select_all()

            # Ctrl+C: Copy
            if pyxel.btnp(pyxel.KEY_C):
                self.copy()

            # Ctrl+X: Cut
            if pyxel.btnp(pyxel.KEY_X):
                self.cut()

            # Ctrl+V: Paste
            if pyxel.btnp(pyxel.KEY_V):
                self.paste()
            return
        with_select_key = pyxel.btn(pyxel.KEY_SHIFT)
        if pyxel.btnp(pyxel.KEY_LEFT, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
            self.move_left(with_select_key)
        if pyxel.btnp(pyxel.KEY_RIGHT, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
            self.move_right(with_select_key)
        if pyxel.btnp(pyxel.KEY_UP, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
            self.move_up(with_select_key)
        if pyxel.btnp(pyxel.KEY_DOWN, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
            self.move_down(with_select_key)
        if pyxel.btnp(pyxel.KEY_BACKSPACE, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
            self.backspace()
        if pyxel.btnp(pyxel.KEY_DELETE, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
            self.delete()



import pyxel

from .settings import INPUT_FIELD_COLOR, INPUT_TEXT_COLOR
from .text_button import TextButton
from .widget import Widget


class NumberPicker(Widget):
    """
    Variables:
        value_var

    Events:
        change (value)
    """

    def __init__(self, parent, x, y, *, min_value, max_value, value, **kwargs):
        self._number_len = max(len(str(min_value)), len(str(max_value)))
        width = self._number_len * 4 + 21
        super().__init__(parent, x, y, width, 7, **kwargs)
        self._min_value = min_value
        self._max_value = max_value

        # Initialize value_var
        self.new_var("value_var", value)
        self.add_var_event_listener("value_var", "set", self.__on_value_set)
        self.add_var_event_listener("value_var", "change", self.__on_value_change)

        # Initialize dec button
        self.dec_button = TextButton(self, 0, 0, text="-")
        self.dec_button.add_event_listener("press", self.__on_dec_button_press)

        # Initialize inc button
        self.inc_button = TextButton(self, self.width - 7, 0, text="+")
        self.inc_button.add_event_listener("press", self.__on_inc_button_press)

        # Set event listeners
        self.add_event_listener("draw", self.__on_draw)

    def __on_value_set(self, value):
        return min(max(value, self._min_value), self._max_value)

    def __on_value_change(self, value):
        self.dec_button.is_enabled_var = self.value_var > self._min_value
        self.inc_button.is_enabled_var = self.value_var < self._max_value
        self.trigger_event("change", value)

    def __on_dec_button_press(self):
        self.value_var -= 10 if pyxel.btn(pyxel.KEY_SHIFT) else 1

    def __on_inc_button_press(self):
        self.value_var += 10 if pyxel.btn(pyxel.KEY_SHIFT) else 1

    def __on_draw(self):
        pyxel.rect(self.x + 9, self.y, self.width - 18, self.height, INPUT_FIELD_COLOR)
        pyxel.text(
            self.x + 11,
            self.y + 1,
            ("{:>" + str(self._number_len) + "}").format(self.value_var),
            INPUT_TEXT_COLOR,
        )



import pyxel

from .button import Button
from .settings import WIDGET_BACKGROUND_COLOR, WIDGET_PANEL_COLOR
from .widget import Widget


class ScrollBar(Widget):
    """
    Variables:
        value_var

    Events:
        change (value)
    """

    def __init__(
        self,
        parent,
        x,
        y,
        *,
        width=None,
        height=None,
        scroll_amount,
        slider_amount,
        value,
        with_shadow=True,
        **kwargs
    ):
        if width is None and height is None or width is not None and height is not None:
            raise ValueError("Either width or height should be specified")
        if height is not None:
            width = 7
            self._is_vertical = True
        else:
            height = 7
            self._is_vertical = False
        super().__init__(parent, x, y, width, height, **kwargs)
        self.scroll_amount = scroll_amount
        self.slider_amount = slider_amount
        self._with_shadow = with_shadow
        self._drag_offset = 0
        self._is_dragged = False

        # Initialize value_var
        self.new_var("value_var", value)
        self.add_var_event_listener("value_var", "set", self.__on_value_set)
        self.add_var_event_listener("value_var", "change", self.__on_value_change)

        # Initialize dec button
        if self._is_vertical:
            btn_w = 7
            btn_h = 6
        else:
            btn_w = 6
            btn_h = 7
        self.dec_button = Button(self, 0, 0, btn_w, btn_h)
        self.dec_button.add_event_listener("press", self.__on_dec_button_press)

        # Initialize inc button
        if self._is_vertical:
            inc_x = 0
            inc_y = height - 6
        else:
            inc_x = width - 6
            inc_y = 0
        self.inc_button = Button(self, inc_x, inc_y, btn_w, btn_h)
        self.inc_button.add_event_listener("press", self.__on_inc_button_press)

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_up", self.__on_mouse_up)
        self.add_event_listener("mouse_drag", self.__on_mouse_drag)
        self.add_event_listener("mouse_repeat", self.__on_mouse_repeat)
        self.add_event_listener("draw", self.__on_draw)

    @property
    def _scroll_size(self):
        return (self.height if self._is_vertical else self.width) - 14

    @property
    def _slider_size(self):
        return round(self._scroll_size * self.slider_amount / self.scroll_amount)

    @property
    def _slider_pos(self):
        return round(7 + self._scroll_size * self.value_var / self.scroll_amount)

    def __on_value_set(self, value):
        return min(max(value, 0), self.scroll_amount)

    def __on_value_change(self, value):
        self.trigger_event("change", value)

    def __on_dec_button_press(self):
        self.value_var = max(self.value_var - 1, 0)

    def __on_inc_button_press(self):
        self.value_var = min(
            self.value_var + 1, self.scroll_amount - self.slider_amount
        )

    def __on_mouse_down(self, key, x, y):
        if key != pyxel.MOUSE_BUTTON_LEFT:
            return
        x -= self.x
        y -= self.y
        self._drag_offset = (y if self._is_vertical else x) - self._slider_pos
        if self._drag_offset < 0:
            self.__on_dec_button_press()
        elif self._drag_offset >= self._slider_size:
            self.__on_inc_button_press()
        else:
            self._is_dragged = True

    def __on_mouse_up(self, key, x, y):
        self._is_dragged = False

    def __on_mouse_drag(self, key, x, y, dx, dy):
        if not self._is_dragged:
            return
        x -= self.x
        y -= self.y
        drag_pos = y if self._is_vertical else x
        value = (
            (drag_pos - self._drag_offset - 6) * self.scroll_amount / self._scroll_size
        )
        self.value_var = int(
            min(max(value, 0), self.scroll_amount - self.slider_amount)
        )

    def __on_mouse_repeat(self, key, x, y):
        if not self._is_dragged:
            self.__on_mouse_down(key, x, y)

    def __on_draw(self):
        x = self.x
        y = self.y
        w = self.width
        h = self.height
        self.draw_panel(x, y, w, h, with_shadow=self._with_shadow)
        inc_col = 6 if self.inc_button.is_pressed_var else WIDGET_BACKGROUND_COLOR
        dec_col = 6 if self.dec_button.is_pressed_var else WIDGET_BACKGROUND_COLOR
        if self._is_vertical:
            # Draw border
            pyxel.rect(x + 1, y + 1, w - 2, 4, dec_col)
            pyxel.rect(x + 1, y + 6, w - 2, h - 12, WIDGET_BACKGROUND_COLOR)
            pyxel.rect(x + 1, y + h - 5, w - 2, 4, inc_col)

            # Draw up arrow
            pyxel.pset(x + 3, y + 2, WIDGET_PANEL_COLOR)
            pyxel.line(x + 2, y + 3, x + w - 3, y + 3, WIDGET_PANEL_COLOR)

            # Draw down arrow
            pyxel.pset(x + 3, y + h - 3, WIDGET_PANEL_COLOR)
            pyxel.line(x + 2, y + h - 4, x + w - 3, y + h - 4, WIDGET_PANEL_COLOR)

            # Draw slider
            pyxel.rect(
                self.x + 2,
                self.y + self._slider_pos,
                3,
                self._slider_size,
                WIDGET_PANEL_COLOR,
            )
        else:
            # Draw border
            pyxel.rect(x + 1, y + 1, 4, h - 2, dec_col)
            pyxel.rect(x + 6, y + 1, w - 12, h - 2, WIDGET_BACKGROUND_COLOR)
            pyxel.rect(x + w - 5, y + 1, 4, h - 2, inc_col)

            # Draw left arrow
            pyxel.pset(x + 2, y + 3, WIDGET_PANEL_COLOR)
            pyxel.line(x + 3, y + 2, x + 3, y + h - 3, WIDGET_PANEL_COLOR)

            # Draw right arrow
            pyxel.pset(x + w - 3, y + h - 4, WIDGET_PANEL_COLOR)
            pyxel.line(x + w - 4, y + 2, x + w - 4, y + h - 3, WIDGET_PANEL_COLOR)

            # Draw slider
            pyxel.rect(
                self.x + self._slider_pos,
                self.y + 2,
                self._slider_size,
                3,
                WIDGET_PANEL_COLOR,
            )



import unittest

import pyxel


class TestPyxel(unittest.TestCase):
    def setUpClass():
        pyxel.init(300, 300, "hoge")

    def test_title(self):
        pyxel.title("hoge")

    def test_colors(self):
        self.assertEqual(pyxel.colors[0], 0)

        pyxel.colors[0] = 0x112233
        self.assertEqual(pyxel.colors[0], 0x112233)

    def test_cls(self):
        pyxel.cls(3)

    def test_input_text(self):
        pyxel.input_text



WIDGET_HOLD_TIME = 16
WIDGET_REPEAT_TIME = 2
WIDGET_CLICK_TIME = 5
WIDGET_CLICK_DIST = 3
WIDGET_PANEL_COLOR = 1
WIDGET_BACKGROUND_COLOR = 7
WIDGET_SHADOW_COLOR = 13

BUTTON_ENABLED_COLOR = 12
BUTTON_DISABLED_COLOR = 5
BUTTON_PRESSED_COLOR = 7
BUTTON_TEXT_COLOR = 1
BUTTON_PRESSING_TIME = 4

INPUT_TEXT_COLOR = 1
INPUT_FIELD_COLOR = 10



import pyxel


def _rect2(self, x1, y1, x2, y2, val):
    x1, x2 = (x1, x2) if x1 < x2 else (x2, x1)
    y1, y2 = (y1, y2) if y1 < y2 else (y2, y1)
    self.rect(x1, y1, x2 - x1 + 1, y2 - y1 + 1, val)


def _rectb2(self, x1, y1, x2, y2, val):
    x1, x2 = (x1, x2) if x1 < x2 else (x2, x1)
    y1, y2 = (y1, y2) if y1 < y2 else (y2, y1)
    self.rectb(x1, y1, x2 - x1 + 1, y2 - y1 + 1, val)


def _elli2(self, x1, y1, x2, y2, val):
    x1, x2 = (x1, x2) if x1 < x2 else (x2, x1)
    y1, y2 = (y1, y2) if y1 < y2 else (y2, y1)
    self.elli(x1, y1, x2 - x1 + 1, y2 - y1 + 1, val)


def _ellib2(self, x1, y1, x2, y2, val):
    x1, x2 = (x1, x2) if x1 < x2 else (x2, x1)
    y1, y2 = (y1, y2) if y1 < y2 else (y2, y1)
    self.ellib(x1, y1, x2 - x1 + 1, y2 - y1 + 1, val)


def _get_slice(self, x, y, width, height):
    data = [[0] * width for _ in range(height)]
    for yi in range(height):
        for xi in range(width):
            data[yi][xi] = self.pget(x + xi, y + yi)
    return data


def _set_slice(self, x, y, slice):
    width = len(slice[0])
    height = len(slice)
    for yi in range(height):
        for xi in range(width):
            self.pset(x + xi, y + yi, slice[yi][xi])


pyxel.Image.rect2 = pyxel.Tilemap.rect2 = _rect2
pyxel.Image.rectb2 = pyxel.Tilemap.rectb2 = _rectb2
pyxel.Image.elli2 = pyxel.Tilemap.elli2 = _elli2
pyxel.Image.ellib2 = pyxel.Tilemap.ellib2 = _ellib2
pyxel.Image.get_slice = pyxel.Tilemap.get_slice = _get_slice
pyxel.Image.set_slice = pyxel.Tilemap.set_slice = _set_slice



import pyxel


def draw_palette(x, y, col):
    rgb = pyxel.colors[col]
    hex = f"#{rgb:06X}"
    dec = f"{rgb >> 16},{(rgb >> 8) & 0xFF},{rgb & 0xFF}"
    pyxel.rect(x, y, 13, 13, col)
    pyxel.text(x + 16, y + 1, hex, 7)
    pyxel.text(x + 16, y + 8, dec, 7)
    pyxel.text(x + 5 - (col // 10) * 2, y + 4, f"{col}", 7 if col < 6 else 0)
    if col == 0:
        pyxel.rectb(x, y, 13, 13, 13)


pyxel.init(255, 81, title="Pyxel Color Palette")
pyxel.cls(0)
for i in range(16):
    draw_palette(2 + (i % 4) * 64, 4 + (i // 4) * 20, i)
pyxel.show()



import pyxel

from .settings import BUTTON_DISABLED_COLOR, BUTTON_ENABLED_COLOR, BUTTON_PRESSED_COLOR
from .widget import Widget


class ToggleButton(Widget):
    """
    Variables:
        is_checked_var

    Events:
        checked
        unchecked
    """

    def __init__(self, parent, x, y, width, height, *, is_checked, **kwargs):
        super().__init__(parent, x, y, width, height, **kwargs)

        # Initialize is_checked_var
        self.new_var("is_checked_var", is_checked)
        self.add_var_event_listener(
            "is_checked_var", "change", self.__on_is_checked_change
        )

        # Initialize event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)

    @property
    def button_color(self):
        if not self.is_enabled_var:
            return BUTTON_DISABLED_COLOR
        elif self.is_checked_var:
            return BUTTON_PRESSED_COLOR
        else:
            return BUTTON_ENABLED_COLOR

    def __on_is_checked_change(self, value):
        self.trigger_event("checked" if value else "unchecked")

    def __on_mouse_down(self, key, x, y):
        if key == pyxel.MOUSE_BUTTON_LEFT:
            self.is_checked_var = not self.is_checked_var



import os

import pyxel

EDITOR_IMAGE_FILE = "assets/editor_220x160.png"
EDITOR_IMAGE = pyxel.Image.from_image(
    os.path.join(os.path.dirname(__file__), EDITOR_IMAGE_FILE)
)

APP_WIDTH = 240
APP_HEIGHT = 180

TOOL_SELECT = 0
TOOL_PENCIL = 1
TOOL_RECTB = 2
TOOL_RECT = 3
TOOL_CIRCB = 4
TOOL_CIRC = 5
TOOL_BUCKET = 6

MAX_SOUND_LENGTH = 48
MAX_MUSIC_LENGTH = 32

TEXT_LABEL_COLOR = 7
HELP_MESSAGE_COLOR = 5

PANEL_FOCUS_COLOR = 7
PANEL_FOCUS_BORDER_COLOR = 0
PANEL_SELECT_FRAME_COLOR = 15
PANEL_SELECT_BORDER_COLOR = 0

PIANO_KEYBOARD_REST_COLOR = 12
PIANO_KEYBOARD_PLAY_COLOR = 14

PIANO_ROLL_CURSOR_PLAY_COLOR = 14
PIANO_ROLL_CURSOR_EDIT_COLOR = 6
PIANO_ROLL_CURSOR_SELECT_COLOR = 15
PIANO_ROLL_BACKGROUND_COLOR = 7
PIANO_ROLL_NOTE_COLOR = 8
PIANO_ROLL_REST_COLOR = 5

OCTAVE_BAR_BACKGROUND_COLOR = 7
OCTAVE_BAR_COLOR = 13

SOUND_FIELD_DATA_NORMAL_COLOR = 1
SOUND_FIELD_DATA_SELECT_COLOR = 7
SOUND_FIELD_CURSOR_EDIT_COLOR = 1
SOUND_FIELD_CURSOR_SELECT_COLOR = 2

MUSIC_FIELD_BACKGROUND_COLOR = 6
MUSIC_FIELD_SOUND_NORMAL_COLOR = 1
MUSIC_FIELD_SOUND_SELECT_COLOR = 7
MUSIC_FIELD_CURSOR_PLAY_COLOR = 8
MUSIC_FIELD_CURSOR_EDIT_COLOR = 1
MUSIC_FIELD_CURSOR_SELECT_COLOR = 2



import pyxel

from .settings import (
    BUTTON_DISABLED_COLOR,
    BUTTON_ENABLED_COLOR,
    BUTTON_PRESSED_COLOR,
    BUTTON_PRESSING_TIME,
)
from .widget import Widget


class Button(Widget):
    """
    Variables:
        is_pressed_var

    Events:
        press
    """

    def __init__(self, parent, x, y, width, height, **kwargs):
        super().__init__(parent, x, y, width, height, **kwargs)
        self._pressing_time = 0

        # Initialize is_pressed_var
        self.new_var("is_pressed_var", None)
        self.add_var_event_listener("is_pressed_var", "get", self.__on_is_pressed_get)
        self.add_var_event_listener("is_pressed_var", "set", self.__on_is_pressed_set)

        # Set event listeners
        self.add_event_listener("mouse_down", self.__on_mouse_down)
        self.add_event_listener("mouse_repeat", self.__on_mouse_down)
        self.add_event_listener("mouse_up", self.__on_mouse_up)
        self.add_event_listener("update", self.__on_update)

    @property
    def button_color(self):
        if not self.is_enabled_var:
            return BUTTON_DISABLED_COLOR
        elif self.is_pressed_var:
            return BUTTON_PRESSED_COLOR
        else:
            return BUTTON_ENABLED_COLOR

    def __on_is_pressed_get(self, value):
        return self._pressing_time > 0

    def __on_is_pressed_set(self, value):
        if value:
            self._pressing_time = BUTTON_PRESSING_TIME + 1
            self.trigger_event("press")
        else:
            self._pressing_time = 0
        return None

    def __on_mouse_down(self, key, x, y):
        if key == pyxel.MOUSE_BUTTON_LEFT:
            self.is_pressed_var = True

    def __on_mouse_up(self, key, x, y):
        if key == pyxel.MOUSE_BUTTON_LEFT:
            self.is_pressed_var = False

    def __on_update(self):
        if self._pressing_time > 0:
            self._pressing_time -= 1



import pyxel

from .settings import (
    WIDGET_CLICK_DIST,
    WIDGET_CLICK_TIME,
    WIDGET_HOLD_TIME,
    WIDGET_PANEL_COLOR,
    WIDGET_REPEAT_TIME,
    WIDGET_SHADOW_COLOR,
)
from .widget_var import WidgetVar


class MouseCaptureInfo:
    widget = None
    key = None
    time = None
    press_pos = None
    last_pos = None


class Widget:
    """
    Variables:
        is_visible_var
        is_enabled_var

    Events:
        show
        hide
        enabled
        disabled
        mouse_down (key, x, y)
        mouse_up (key, x, y)
        mouse_drag (key, x, y, dx, dy)
        mouse_repeat (key, x, y)
        mouse_click (key, x, y)
        mouse_hover (x, y)
        update
        draw
    """

    _mouse_capture_info = MouseCaptureInfo()

    def __init__(
        self, parent, x, y, width, height, *, is_visible=True, is_enabled=True
    ):
        if parent:
            parent._children.append(self)
        self._parent = parent
        self._children = []
        self._x = x
        self._y = y
        self._width = width
        self._height = height
        self._event_listeners = {}

        # Initialize is_visible_var
        self.new_var("is_visible_var", is_visible)
        self.add_var_event_listener("is_visible_var", "get", self.__on_is_visible_get)
        self.add_var_event_listener(
            "is_visible_var", "change", self.__on_is_visible_change
        )

        # Initialize is_enabled_var
        self.new_var("is_enabled_var", is_enabled)
        self.add_var_event_listener("is_enabled_var", "get", self.__on_is_enabled_get)
        self.add_var_event_listener(
            "is_enabled_var", "change", self.__on_is_enabled_change
        )

    @property
    def x(self):
        return (self._parent.x + self._x) if self._parent else self._x

    @property
    def y(self):
        return (self._parent.y + self._y) if self._parent else self._y

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    def is_hit(self, x, y):
        x -= self.x
        y -= self.y
        return 0 <= x <= self.width - 1 and 0 <= y <= self.height - 1

    def set_pos(self, x, y):
        self._x = x
        self._y = y

    def set_size(self, width, height):
        self._width = width
        self._height = height

    def add_event_listener(self, event, listener):
        self._event_listeners.setdefault(event, [])
        self._event_listeners[event].append(listener)

    def remove_event_listener(self, event, listener):
        self._event_listeners.setdefault(event, [])
        self._event_listeners[event].remove(listener)

    def trigger_event(self, event, *args):
        self._event_listeners.setdefault(event, [])
        for listener in self._event_listeners[event]:
            listener(*args)

    def update_all(self):
        capture_widget = Widget._mouse_capture_info.widget
        if capture_widget:
            capture_widget._process_capture()
        else:
            self._process_input()
        self._update()

    def _process_input(self):
        if not self.is_visible_var or not self.is_enabled_var:
            return False
        for widget in reversed(self._children):
            if widget._process_input():
                return True
        x = pyxel.mouse_x
        y = pyxel.mouse_y
        if self.is_hit(x, y):
            if pyxel.btnp(pyxel.MOUSE_BUTTON_LEFT):
                key = pyxel.MOUSE_BUTTON_LEFT
            elif pyxel.btnp(pyxel.MOUSE_BUTTON_RIGHT):
                key = pyxel.MOUSE_BUTTON_RIGHT
            elif pyxel.btnp(pyxel.MOUSE_BUTTON_MIDDLE):
                key = pyxel.MOUSE_BUTTON_MIDDLE
            else:
                key = None
            if key is not None:
                self._start_capture(key)
                self.trigger_event("mouse_down", key, x, y)
            self.trigger_event("mouse_hover", x, y)
            return True
        return False

    def _start_capture(self, key):
        capture_info = Widget._mouse_capture_info
        capture_info.widget = self
        capture_info.key = key
        capture_info.time = pyxel.frame_count
        capture_info.press_pos = (pyxel.mouse_x, pyxel.mouse_y)
        capture_info.last_pos = capture_info.press_pos

    def _end_capture(self):
        capture_info = Widget._mouse_capture_info
        capture_info.widget = None
        capture_info.key = None
        capture_info.time = None
        capture_info.press_pos = None
        capture_info.last_pos = None

    def _process_capture(self):
        capture_info = Widget._mouse_capture_info
        last_x, last_y = capture_info.last_pos
        x = pyxel.mouse_x
        y = pyxel.mouse_y
        if x != last_x or y != last_y:
            self.trigger_event(
                "mouse_drag",
                capture_info.key,
                x,
                y,
                x - last_x,
                y - last_y,
            )
            capture_info.last_pos = (x, y)

        if self.is_hit(x, y):
            self.trigger_event("mouse_hover", x, y)

        if pyxel.btnp(capture_info.key, WIDGET_HOLD_TIME, WIDGET_REPEAT_TIME):
            self.trigger_event("mouse_repeat", capture_info.key, x, y)

        if pyxel.btnr(capture_info.key):
            self.trigger_event("mouse_up", capture_info.key, x, y)
            press_x, press_y = capture_info.press_pos
            if (
                pyxel.frame_count <= capture_info.time + WIDGET_CLICK_TIME
                and abs(x - press_x) <= WIDGET_CLICK_DIST
                and abs(y - press_y) <= WIDGET_CLICK_DIST
            ):
                self.trigger_event("mouse_click", capture_info.key, x, y)
            self._end_capture()

    def _update(self):
        if not self.is_visible_var:
            return
        self.trigger_event("update")
        for child in self._children:
            child._update()

    def draw_all(self):
        if not self.is_visible_var:
            return
        self.trigger_event("draw")
        for child in self._children:
            child.draw_all()

    @staticmethod
    def draw_panel(x, y, width, height, *, with_shadow=True):
        w = width
        h = height
        pyxel.line(x + 1, y, x + w - 2, y, WIDGET_PANEL_COLOR)
        pyxel.rect(x, y + 1, w, h - 2, WIDGET_PANEL_COLOR)
        pyxel.line(x + 1, y + h - 1, x + w - 2, y + h - 1, WIDGET_PANEL_COLOR)
        if with_shadow:
            pyxel.line(x + 2, y + h, x + w - 1, y + h, WIDGET_SHADOW_COLOR)
            pyxel.line(x + w, y + 2, x + w, y + h - 1, WIDGET_SHADOW_COLOR)
            pyxel.pset(x + w - 1, y + h - 1, WIDGET_SHADOW_COLOR)

    def new_var(self, name, value):
        member_name = self._widget_var_name(name)
        widget_var = WidgetVar(value)
        setattr(self, member_name, widget_var)

        def getter(self):
            return getattr(self, member_name).get()

        def setter(self, value):
            getattr(self, member_name).set(value)

        setattr(self.__class__, name, property(getter, setter))

    def copy_var(self, name, src_widget, src_name=None):
        member_name = self._widget_var_name(name)
        src_member_name = self._widget_var_name(src_name or name)
        widget_var = getattr(src_widget, src_member_name)
        setattr(self, member_name, widget_var)

        def getter(self):
            return getattr(self, member_name).get()

        def setter(self, value):
            getattr(self, member_name).set(value)

        setattr(self.__class__, name, property(getter, setter))

    def add_var_event_listener(self, name, event, listener):
        member_name = self._widget_var_name(name)
        widget_var = getattr(self, member_name)
        widget_var.add_event_listener(event, listener)

    def remove_var_event_listener(self, name, event, listener):
        member_name = self._widget_var_name(name)
        widget_var = getattr(self, member_name)
        widget_var.remove_event_listener(event, listener)

    @staticmethod
    def _widget_var_name(name):
        return "_widget_var_" + name

    def __on_is_visible_get(self, value):
        return (self._parent.is_visible_var and value) if self._parent else value

    def __on_is_visible_change(self, value):
        self._trigger_visible_event(value)

    def _trigger_visible_event(self, is_visible):
        self.trigger_event("show" if is_visible else "hide")
        for child in self._children:
            if child.is_visible_var == is_visible:
                child._trigger_visible_event(is_visible)

    def __on_is_enabled_get(self, value):
        return (self._parent.is_enabled_var and value) if self._parent else value

    def __on_is_enabled_change(self, value):
        self._trigger_enabled_event(value)

    def _trigger_enabled_event(self, is_enabled):
        self.trigger_event("enabled" if is_enabled else "disabled")
        for child in self._children:
            if child.is_enabled_var == is_enabled:
                child._trigger_enabled_event(is_enabled)



import pyxel


class App:
    def __init__(self):
        pyxel.init(200, 150, title="Pyxel Triangle API")
        self.triangles = [(100, 24, 7, 143, 193, 143, 7)]
        pyxel.cls(13)
        pyxel.text(6, 6, "tri(x1,y1,x2,y2,x3,y3,col)", 7)
        pyxel.text(6, 14, "trib(x1,y1,x2,y2,x3,y3,col)", 7)
        pyxel.run(self.update, self.draw)

    def update(self):
        if pyxel.btnp(pyxel.KEY_Q):
            pyxel.quit()

    def draw(self):
        if self.triangles:
            triangle = self.triangles.pop(0)
            self.draw_triangle(*triangle)

    def draw_triangle(self, x1, y1, x2, y2, x3, y3, n):
        if n == 0:
            return
        col = n + 7
        if n % 2 == 0:
            pyxel.tri(x1, y1, x2, y2, x3, y3, col)
        else:
            pyxel.trib(x1, y1, x2, y2, x3, y3, col)
        h1 = (x1 + x2) / 2
        w1 = (y1 + y2) / 2
        h2 = (x2 + x3) / 2
        w2 = (y2 + y3) / 2
        h3 = (x3 + x1) / 2
        w3 = (y3 + y1) / 2
        self.triangles.append((x1, y1, h1, w1, h3, w3, n - 1))
        self.triangles.append((h1, w1, x2, y2, h2, w2, n - 1))
        self.triangles.append((h3, w3, h2, w2, x3, y3, n - 1))


App()



import pyxel

from .canvas_panel import CanvasPanel
from .editor_base import EditorBase
from .image_viewer import ImageViewer
from .settings import EDITOR_IMAGE, TEXT_LABEL_COLOR, TOOL_PENCIL
from .tilemap_viewer import TilemapViewer
from .widgets import NumberPicker, RadioButton


class TilemapEditor(EditorBase):
    """
    Variables:
        color_var
        tool_var
        image_no_var
        canvas_var
        focus_x_var
        focus_y_var

        tilemap_no_var
        tile_x_var
        tile_y_var
        tile_w_var
        tile_h_var

    Events:
        undo (data)
        redo (data)
        drop (filename)
    """

    def __init__(self, parent):
        super().__init__(parent)

        # Initialize canvas_var
        self.new_var("canvas_var", None)
        self.add_var_event_listener("canvas_var", "get", self.__on_canvas_get)

        # Initialize color_var
        self.new_var("color_var", (255, 255))

        # Initialize tool button
        self._tool_button = RadioButton(
            self,
            81,
            161,
            img=EDITOR_IMAGE,
            u=63,
            v=0,
            num_buttons=7,
            value=TOOL_PENCIL,
        )
        self.add_tool_button_help(self._tool_button)
        self.copy_var("tool_var", self._tool_button, "value_var")

        # Initialize tilemap picker
        self._tilemap_picker = NumberPicker(
            self, 48, 161, min_value=0, max_value=pyxel.NUM_TILEMAPS - 1, value=0
        )
        self._tilemap_picker.add_event_listener(
            "change", self.__on_tilemap_picker_change
        )
        self.add_number_picker_help(self._tilemap_picker)
        self.copy_var("tilemap_no_var", self._tilemap_picker, "value_var")

        # Initialize tilemap viewer
        self._tilemap_viewer = TilemapViewer(self)
        self.copy_var("focus_x_var", self._tilemap_viewer, "focus_x_var")
        self.copy_var("focus_y_var", self._tilemap_viewer, "focus_y_var")

        # Initialize image picker
        self._image_picker = NumberPicker(
            self,
            192,
            161,
            min_value=0,
            max_value=pyxel.NUM_IMAGES - 1,
            value=pyxel.tilemap(self.tilemap_no_var).refimg,
        )
        self._image_picker.add_event_listener("change", self.__on_image_picker_change)
        self.add_number_picker_help(self._image_picker)
        self.copy_var("image_no_var", self._image_picker, "value_var")

        # Initialize image viewer
        self._image_viewer = ImageViewer(self)
        self.copy_var("tile_x_var", self._image_viewer, "focus_x_var")
        self.copy_var("tile_y_var", self._image_viewer, "focus_y_var")
        self.copy_var("tile_w_var", self._image_viewer, "focus_w_var")
        self.copy_var("tile_h_var", self._image_viewer, "focus_h_var")

        # Initialize canvas panel
        self._canvas_panel = CanvasPanel(self)

        # Set event listeners
        self.add_event_listener("undo", self.__on_undo)
        self.add_event_listener("redo", self.__on_redo)
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

    def __on_canvas_get(self, value):
        return pyxel.tilemap(self.tilemap_no_var)

    def __on_tilemap_picker_change(self, value):
        self.image_no_var = pyxel.tilemap(value).refimg

    def __on_image_picker_change(self, value):
        pyxel.tilemap(self.tilemap_no_var).refimg = value

    def __on_undo(self, data):
        self.tilemap_no_var = data["tilemap_no"]
        self.focus_x_var, self.focus_y_var = data["focus_pos"]
        self.canvas_var.set_slice(
            self.focus_x_var * 8, self.focus_y_var * 8, data["old_canvas"]
        )

    def __on_redo(self, data):
        self.tilemap_no_var = data["tilemap_no"]
        self.focus_x_var, self.focus_y_var = data["focus_pos"]
        self.canvas_var.set_slice(
            self.focus_x_var * 8, self.focus_y_var * 8, data["new_canvas"]
        )

    def __on_update(self):
        self.check_tool_button_shortcuts()

    def __on_draw(self):
        self.draw_panel(11, 156, 136, 17)
        self.draw_panel(157, 156, 72, 17)
        pyxel.text(18, 162, "TILEMAP", TEXT_LABEL_COLOR)
        pyxel.text(18, 162, "TILEMAP", TEXT_LABEL_COLOR)
        pyxel.text(170, 162, "IMAGE", TEXT_LABEL_COLOR)



import pyxel

from .editor_base import EditorBase
from .field_cursor import FieldCursor
from .octave_bar import OctaveBar
from .piano_keyboard import PianoKeyboard
from .piano_roll import PianoRoll
from .settings import EDITOR_IMAGE, MAX_SOUND_LENGTH, TEXT_LABEL_COLOR
from .sound_field import SoundField
from .widgets import ImageButton, ImageToggleButton, NumberPicker


class SoundEditor(EditorBase):
    """
    Variables:
        sound_no_var
        speed_var
        octave_var
        note_var
        is_playing_var
    """

    def __init__(self, parent):
        super().__init__(parent)
        self._history_data = None
        self.copy_var("help_message_var", parent)

        # Initialize field cursor
        self.field_cursor = FieldCursor(
            max_field_length=MAX_SOUND_LENGTH,
            field_wrap_length=MAX_SOUND_LENGTH,
            get_field=self.get_field,
            add_pre_history=self.add_pre_history,
            add_post_history=self.add_post_history,
            cross_filed_copying=False,
        )

        # Initialize octave_var
        self.new_var("octave_var", 2)

        # Initialize is_playing_var
        self.new_var("is_playing_var", None)
        self.add_var_event_listener(
            "is_playing_var", "get", self.__on_is_playing_var_get
        )

        # Initialize sound picker
        self._sound_picker = NumberPicker(
            self, 45, 17, min_value=0, max_value=pyxel.NUM_SOUNDS - 1, value=0
        )
        self._sound_picker.add_event_listener("change", self.__on_sound_picker_change)
        self.add_number_picker_help(self._sound_picker)
        self.copy_var("sound_no_var", self._sound_picker, "value_var")

        # Initialize speed picker
        self._speed_picker = NumberPicker(
            self, 105, 17, min_value=1, max_value=99, value=pyxel.sound(0).speed
        )
        self._speed_picker.add_event_listener("change", self.__on_speed_picker_change)
        self.add_number_picker_help(self._speed_picker)
        self.copy_var("speed_var", self._speed_picker, "value_var")

        # Initialize play button
        self._play_button = ImageButton(self, 185, 17, img=EDITOR_IMAGE, u=126, v=0)
        self._play_button.add_event_listener("press", self.__on_play_button_press)
        self._play_button.add_event_listener(
            "mouse_hover", self.__on_play_button_mouse_hover
        )

        # Initialize stop button
        self._stop_button = ImageButton(
            self, 195, 17, img=EDITOR_IMAGE, u=135, v=0, is_enabled=False
        )
        self._stop_button.add_event_listener("press", self.__on_stop_button_press)
        self._stop_button.add_event_listener(
            "mouse_hover", self.__on_stop_button_mouse_hover
        )

        # Initialize loop button
        self._loop_button = ImageToggleButton(
            self, 205, 17, img=EDITOR_IMAGE, u=144, v=0, is_checked=False
        )
        self._loop_button.add_event_listener(
            "mouse_hover", self.__on_loop_button_mouse_hover
        )
        self.copy_var("should_loop_var", self._loop_button, "is_checked_var")

        # Initialize piano keyboard
        self._piano_keyboard = PianoKeyboard(self)
        self.copy_var("note_var", self._piano_keyboard)

        # Initialize piano roll
        self._piano_roll = PianoRoll(self)

        # Initialize sound field
        self._sound_field = SoundField(self)

        # Initialize left octave bar
        self._left_octave_bar = OctaveBar(self, 12, 25)

        # Initialize right octave bar
        self._right_octave_bar = OctaveBar(self, 224, 25)

        # Set event listeners
        self.add_event_listener("undo", self.__on_undo)
        self.add_event_listener("redo", self.__on_redo)
        self.add_event_listener("hide", self.__on_hide)
        self.add_event_listener("update", self.__on_update)
        self.add_event_listener("draw", self.__on_draw)

    @property
    def keyboard_note(self):
        return self._piano_keyboard.note

    def get_field(self, index):
        sound = pyxel.sound(self.sound_no_var)
        if index == 0:
            return sound.notes
        elif index == 1:
            return sound.tones
        elif index == 2:
            return sound.volumes
        elif index == 3:
            return sound.effects
        else:
            return None

    def add_pre_history(self, x, y):
        self._history_data = data = {}
        data["sound_no"] = self.sound_no_var
        data["old_cursor_pos"] = (x, y)
        data["old_field"] = self.field_cursor.field.to_list()

    def add_post_history(self, x, y):
        data = self._history_data
        data["new_cursor_pos"] = (x, y)
        data["new_field"] = self.field_cursor.field.to_list()
        if data["old_field"] != data["new_field"]:
            self.add_history(self._history_data)

    def get_field_help_message(self):
        cursor_y = self.field_cursor.y
        if cursor_y == 0:
            return "NOTE:CLICK/PIANO_KEY+ENTER/BS/DEL"
        elif cursor_y == 1:
            return "TONE:T/S/P/N/BS/DEL"
        elif cursor_y == 2:
            return "VOLUME:0-7/BS/DEL"
        elif cursor_y == 3:
            return "EFFECT:N/S/V/F/BS/DEL"
        else:
            return ""

    def _play(self, is_partial):
        self._sound_picker.is_enabled_var = False
        self._speed_picker.is_enabled_var = False
        self._play_button.is_enabled_var = False
        self._stop_button.is_enabled_var = True
        self._loop_button.is_enabled_var = False
        tick = self.field_cursor.x * self.speed_var if is_partial else None
        pyxel.play(0, self.sound_no_var, tick, loop=self.should_loop_var)

    def _stop(self):
        self._sound_picker.is_enabled_var = True
        self._speed_picker.is_enabled_var = True
        self._play_button.is_enabled_var = True
        self._stop_button.is_enabled_var = False
        self._loop_button.is_enabled_var = True
        pyxel.stop(0)

    def __on_is_playing_var_get(self, value):
        return pyxel.play_pos(0) is not None

    def __on_sound_picker_change(self, value):
        sound = pyxel.sound(value)
        self._speed_picker.value = sound.speed

    def __on_speed_picker_change(self, value):
        sound = pyxel.sound(self.sound_no_var)
        sound.speed = value

    def __on_play_button_press(self):
        self._play(pyxel.btn(pyxel.KEY_SHIFT))

    def __on_play_button_mouse_hover(self, x, y):
        self.help_message_var = "PLAY:SPACE PART-PLAY:SHIFT+SPACE"

    def __on_stop_button_press(self):
        self._stop()

    def __on_stop_button_mouse_hover(self, x, y):
        self.help_message_var = "STOP:SPACE"

    def __on_loop_button_mouse_hover(self, x, y):
        self.help_message_var = "LOOP:L"

    def __on_undo(self, data):
        self._stop()
        self.sound_no_var = data["sound_no"]
        self.field_cursor.move_to(*data["old_cursor_pos"], False)
        self.field_cursor.field.from_list(data["old_field"])

    def __on_redo(self, data):
        self._stop()
        self.sound_no_var = data["sound_no"]
        self.field_cursor.move_to(*data["new_cursor_pos"], False)
        self.field_cursor.field.from_list(data["new_field"])

    def __on_hide(self):
        self._stop()

    def __on_update(self):
        sound = pyxel.sound(self.sound_no_var)
        if self.speed_var != sound.speed:
            self.speed_var = sound.speed
        if pyxel.btnp(pyxel.KEY_SPACE):
            if self.is_playing_var:
                self._stop_button.is_pressed_var = True
                return
            else:
                self._play_button.is_pressed_var = True
        if not self._play_button.is_enabled_var and not self.is_playing_var:
            self._stop()
        if self._loop_button.is_enabled_var and pyxel.btnp(pyxel.KEY_L):
            self.should_loop_var = not self.should_loop_var
        if pyxel.btnp(pyxel.KEY_PAGEUP):
            self.octave_var = min(self.octave_var + 1, 3)
        if pyxel.btnp(pyxel.KEY_PAGEDOWN):
            self.octave_var = max(self.octave_var - 1, 0)
        self.field_cursor.process_input()

    def __on_draw(self):
        self.draw_panel(11, 16, 218, 157)
        pyxel.text(23, 18, "SOUND", TEXT_LABEL_COLOR)
        pyxel.text(83, 18, "SPEED", TEXT_LABEL_COLOR)



import pyxel

SCENE_TITLE = 0
SCENE_PLAY = 1
SCENE_GAMEOVER = 2

NUM_STARS = 100
STAR_COLOR_HIGH = 12
STAR_COLOR_LOW = 5

PLAYER_WIDTH = 8
PLAYER_HEIGHT = 8
PLAYER_SPEED = 2

BULLET_WIDTH = 2
BULLET_HEIGHT = 8
BULLET_COLOR = 11
BULLET_SPEED = 4

ENEMY_WIDTH = 8
ENEMY_HEIGHT = 8
ENEMY_SPEED = 1.5

BLAST_START_RADIUS = 1
BLAST_END_RADIUS = 8
BLAST_COLOR_IN = 7
BLAST_COLOR_OUT = 10

enemies = []
bullets = []
blasts = []


def update_list(list):
    for elem in list:
        elem.update()


def draw_list(list):
    for elem in list:
        elem.draw()


def cleanup_list(list):
    i = 0
    while i < len(list):
        elem = list[i]
        if not elem.is_alive:
            list.pop(i)
        else:
            i += 1


class Background:
    def __init__(self):
        self.stars = []
        for i in range(NUM_STARS):
            self.stars.append(
                (
                    pyxel.rndi(0, pyxel.width - 1),
                    pyxel.rndi(0, pyxel.height - 1),
                    pyxel.rndf(1, 2.5),
                )
            )

    def update(self):
        for i, (x, y, speed) in enumerate(self.stars):
            y += speed
            if y >= pyxel.height:
                y -= pyxel.height
            self.stars[i] = (x, y, speed)

    def draw(self):
        for (x, y, speed) in self.stars:
            pyxel.pset(x, y, STAR_COLOR_HIGH if speed > 1.8 else STAR_COLOR_LOW)


class Player:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.w = PLAYER_WIDTH
        self.h = PLAYER_HEIGHT
        self.is_alive = True

    def update(self):
        if pyxel.btn(pyxel.KEY_LEFT):
            self.x -= PLAYER_SPEED
        if pyxel.btn(pyxel.KEY_RIGHT):
            self.x += PLAYER_SPEED
        if pyxel.btn(pyxel.KEY_UP):
            self.y -= PLAYER_SPEED
        if pyxel.btn(pyxel.KEY_DOWN):
            self.y += PLAYER_SPEED
        self.x = max(self.x, 0)
        self.x = min(self.x, pyxel.width - self.w)
        self.y = max(self.y, 0)
        self.y = min(self.y, pyxel.height - self.h)

        if pyxel.btnp(pyxel.KEY_SPACE):
            Bullet(
                self.x + (PLAYER_WIDTH - BULLET_WIDTH) / 2, self.y - BULLET_HEIGHT / 2
            )
            pyxel.play(0, 0)

    def draw(self):
        pyxel.blt(self.x, self.y, 0, 0, 0, self.w, self.h, 0)


class Bullet:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.w = BULLET_WIDTH
        self.h = BULLET_HEIGHT
        self.is_alive = True
        bullets.append(self)

    def update(self):
        self.y -= BULLET_SPEED
        if self.y + self.h - 1 < 0:
            self.is_alive = False

    def draw(self):
        pyxel.rect(self.x, self.y, self.w, self.h, BULLET_COLOR)


class Enemy:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.w = ENEMY_WIDTH
        self.h = ENEMY_HEIGHT
        self.dir = 1
        self.timer_offset = pyxel.rndi(0, 59)
        self.is_alive = True
        enemies.append(self)

    def update(self):
        if (pyxel.frame_count + self.timer_offset) % 60 < 30:
            self.x += ENEMY_SPEED
            self.dir = 1
        else:
            self.x -= ENEMY_SPEED
            self.dir = -1
        self.y += ENEMY_SPEED
        if self.y > pyxel.height - 1:
            self.is_alive = False

    def draw(self):
        pyxel.blt(self.x, self.y, 0, 8, 0, self.w * self.dir, self.h, 0)


class Blast:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.radius = BLAST_START_RADIUS
        self.is_alive = True
        blasts.append(self)

    def update(self):
        self.radius += 1
        if self.radius > BLAST_END_RADIUS:
            self.is_alive = False

    def draw(self):
        pyxel.circ(self.x, self.y, self.radius, BLAST_COLOR_IN)
        pyxel.circb(self.x, self.y, self.radius, BLAST_COLOR_OUT)


class App:
    def __init__(self):
        pyxel.init(120, 160, title="Pyxel Shooter")
        pyxel.image(0).set(
            0,
            0,
            [
                "00c00c00",
                "0c7007c0",
                "0c7007c0",
                "c703b07c",
                "77033077",
                "785cc587",
                "85c77c58",
                "0c0880c0",
            ],
        )
        pyxel.image(0).set(
            8,
            0,
            [
                "00088000",
                "00ee1200",
                "08e2b180",
                "02882820",
                "00222200",
                "00012280",
                "08208008",
                "80008000",
            ],
        )
        pyxel.sound(0).set("a3a2c1a1", "p", "7", "s", 5)
        pyxel.sound(1).set("a3a2c2c2", "n", "7742", "s", 10)
        self.scene = SCENE_TITLE
        self.score = 0
        self.background = Background()
        self.player = Player(pyxel.width / 2, pyxel.height - 20)
        pyxel.run(self.update, self.draw)

    def update(self):
        if pyxel.btn(pyxel.KEY_Q):
            pyxel.quit()

        self.background.update()
        if self.scene == SCENE_TITLE:
            self.update_title_scene()
        elif self.scene == SCENE_PLAY:
            self.update_play_scene()
        elif self.scene == SCENE_GAMEOVER:
            self.update_gameover_scene()

    def update_title_scene(self):
        if pyxel.btnp(pyxel.KEY_RETURN):
            self.scene = SCENE_PLAY

    def update_play_scene(self):
        if pyxel.frame_count % 6 == 0:
            Enemy(pyxel.rndi(0, pyxel.width - ENEMY_WIDTH), 0)

        for enemy in enemies:
            for bullet in bullets:
                if (
                    enemy.x + enemy.w > bullet.x
                    and bullet.x + bullet.w > enemy.x
                    and enemy.y + enemy.h > bullet.y
                    and bullet.y + bullet.h > enemy.y
                ):
                    enemy.is_alive = False
                    bullet.is_alive = False
                    blasts.append(
                        Blast(enemy.x + ENEMY_WIDTH / 2, enemy.y + ENEMY_HEIGHT / 2)
                    )
                    pyxel.play(1, 1)
                    self.score += 10

        for enemy in enemies:
            if (
                self.player.x + self.player.w > enemy.x
                and enemy.x + enemy.w > self.player.x
                and self.player.y + self.player.h > enemy.y
                and enemy.y + enemy.h > self.player.y
            ):
                enemy.is_alive = False
                blasts.append(
                    Blast(
                        self.player.x + PLAYER_WIDTH / 2,
                        self.player.y + PLAYER_HEIGHT / 2,
                    )
                )
                pyxel.play(1, 1)
                self.scene = SCENE_GAMEOVER

        self.player.update()
        update_list(bullets)
        update_list(enemies)
        update_list(blasts)
        cleanup_list(enemies)
        cleanup_list(bullets)
        cleanup_list(blasts)

    def update_gameover_scene(self):
        update_list(bullets)
        update_list(enemies)
        update_list(blasts)
        cleanup_list(enemies)
        cleanup_list(bullets)
        cleanup_list(blasts)

        if pyxel.btnp(pyxel.KEY_RETURN):
            self.scene = SCENE_PLAY
            self.player.x = pyxel.width / 2
            self.player.y = pyxel.height - 20
            self.score = 0
            enemies.clear()
            bullets.clear()
            blasts.clear()

    def draw(self):
        pyxel.cls(0)
        self.background.draw()
        if self.scene == SCENE_TITLE:
            self.draw_title_scene()
        elif self.scene == SCENE_PLAY:
            self.draw_play_scene()
        elif self.scene == SCENE_GAMEOVER:
            self.draw_gameover_scene()
        pyxel.text(39, 4, f"SCORE {self.score:5}", 7)

    def draw_title_scene(self):
        pyxel.text(35, 66, "Pyxel Shooter", pyxel.frame_count % 16)
        pyxel.text(31, 126, "- PRESS ENTER -", 13)

    def draw_play_scene(self):
        self.player.draw()
        draw_list(bullets)
        draw_list(enemies)
        draw_list(blasts)

    def draw_gameover_scene(self):
        draw_list(bullets)
        draw_list(enemies)
        draw_list(blasts)
        pyxel.text(43, 66, "GAME OVER", 8)
        pyxel.text(31, 126, "- PRESS ENTER -", 13)


App()



import pyxel

from .button import Button
from .settings import BUTTON_ENABLED_COLOR


class ImageButton(Button):
    """
    Variables:
        is_pressed_var

    Events:
        press
    """

    def __init__(self, parent, x, y, *, img, u, v, **kwargs):
        super().__init__(parent, x, y, 7, 7, **kwargs)
        self._img = img
        self._u = u
        self._v = v

        # Set event listeners
        self.add_event_listener("draw", self.__on_draw)

    def __on_draw(self):
        pyxel.pal(BUTTON_ENABLED_COLOR, self.button_color)
        pyxel.blt(
            self.x,
            self.y,
            self._img,
            self._u,
            self._v,
            self.width,
            self.height,
            0,
        )
        pyxel.pal()



from .button import Button  # noqa: F401
from .color_picker import ColorPicker  # noqa: F401
from .image_button import ImageButton  # noqa: F401
from .image_toggle_button import ImageToggleButton  # noqa: F401
from .number_picker import NumberPicker  # noqa: F401
from .radio_button import RadioButton  # noqa: F401
from .scroll_bar import ScrollBar  # noqa: F401
from .text_button import TextButton  # noqa: F401
from .toggle_button import ToggleButton  # noqa: F401
from .widget import Widget  # noqa: F401
